from __future__ import annotations

from collections.abc import Callable
from enum import Enum
from types import EllipsisType, FunctionType, MethodType, ModuleType, NoneType, NotImplementedType, UnionType
from typing import TYPE_CHECKING, Annotated, Any, Literal, TypeVar, Union, get_origin, overload

if TYPE_CHECKING:
    from sonolus.script.internal.value import Value


@overload
def meta_fn[T: Callable](fn: T) -> T: ...


@overload
def meta_fn[T: Callable]() -> Callable[[T], T]: ...


def meta_fn(fn=None):
    """Marks a function as a meta function to be called directly without the AST visitor.

    This can also improve performance in some cases by avoiding the overhead of the AST visitor.
    """

    # noinspection PyShadowingNames
    def decorator(fn):
        fn._meta_fn_ = True
        return fn

    if fn is None:
        return decorator
    return decorator(fn)


def validate_value[T](value: T) -> Value | T:
    result = try_validate_value(value)
    if result is None:
        raise TypeError(f"Unsupported value: {value!r}")
    return result


def try_validate_value(value: Any) -> Value | None:
    from sonolus.script.globals import _GlobalPlaceholder
    from sonolus.script.internal.constant import BasicConstantValue, TypingSpecialFormConstant
    from sonolus.script.internal.dict_impl import DictImpl
    from sonolus.script.internal.generic import PartialGeneric
    from sonolus.script.internal.tuple_impl import TupleImpl
    from sonolus.script.internal.value import Value
    from sonolus.script.num import Num

    try:
        # Unfortunately this is called during import, so this may fail
        from sonolus.script.internal.builtin_impls import BUILTIN_IMPLS

        if id(value) in BUILTIN_IMPLS:
            return validate_value(BUILTIN_IMPLS[id(value)])
    except ImportError:
        pass

    match value:
        case Enum():
            return validate_value(value.value)
        case Value():
            return value
        case type():
            if value in {int, float, bool}:
                return BasicConstantValue.of(Num)
            return BasicConstantValue.of(value)
        case int() | float() | bool():
            return Num._accept_(value)
        case tuple():
            return TupleImpl._accept_(value)
        case dict():
            return DictImpl._accept_(value)
        case (
            PartialGeneric()
            | TypeVar()
            | FunctionType()
            | MethodType()
            | str()
            | ModuleType()
            | NoneType()
            | NotImplementedType()
            | EllipsisType()
            | super()
        ):
            return BasicConstantValue.of(value)
        case special_form if value in {
            Literal,
            Annotated,
            Union,
        }:
            return TypingSpecialFormConstant.of(special_form)
        case other_type if get_origin(value) in {Literal, Annotated, UnionType, tuple, type}:
            return BasicConstantValue.of(other_type)
        case _GlobalPlaceholder():
            return value.get()
        case comptime_value if getattr(comptime_value, "_is_comptime_value_", False):
            return BasicConstantValue.of(comptime_value)
        case _:
            return None



# z-index constants for ordering sprites.

LAYER_STAGE = 0
LAYER_LANE = 1
LAYER_JUDGE_LINE = 2

LAYER_CONNECTOR = 10

LAYER_PREVIEW_COVER = 20
LAYER_MEASURE_LINE = 21
LAYER_SIM_LINE = 22
LAYER_TIME_LINE = 23
LAYER_BPM_CHANGE_LINE = 24
LAYER_TIMESCALE_CHANGE_LINE = 25

LAYER_NOTE_HEAD = 30
LAYER_NOTE = 31
LAYER_ARROW = 32


def get_z(layer: int, lane: float = 0, y: float = 0) -> float:
    """Calculate z-index based on layer, lane, and y-coordinate.

    Lane and y are used to prevent z-fighting between sprites in the same layer.
    """
    return layer * 10000 + lane * 100 + y



# ruff: noqa: B905, C417

from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.containers import VarArray
from sonolus.script.debug import assert_true, debug_log
from tests.script.conftest import run_and_validate

ints = st.integers(min_value=-999, max_value=999)
lists = st.lists(ints, min_size=1, max_size=20)
sets = st.sets(ints, min_size=1, max_size=20)


@st.composite
def list_with_duplicates(draw):
    size = draw(st.integers(min_value=1, max_value=20))
    base_values = draw(st.lists(st.integers(min_value=-10, max_value=10), min_size=size // 2, max_size=size // 2))

    duplicates = draw(
        st.lists(
            st.sampled_from(base_values) if base_values else st.integers(-10, 10),
            min_size=size - len(base_values),
            max_size=size - len(base_values),
        )
    )

    values = base_values + duplicates
    values = draw(st.permutations(values))

    target = draw(st.sampled_from(values))
    return values, target


@st.composite
def list_and_maybe_missing_value(draw):
    size = draw(st.integers(min_value=1, max_value=20))
    values = draw(st.lists(st.integers(min_value=-10, max_value=10), min_size=size, max_size=size))

    if draw(st.booleans()):
        target = draw(st.integers(min_value=-20, max_value=20))
    else:
        target = draw(st.sampled_from(values))

    return values, target


@st.composite
def list_and_index(draw):
    values = draw(lists)
    index = draw(st.integers(min_value=-len(values), max_value=len(values) - 1))
    return values, index


@st.composite
def list_and_insert_index(draw):
    values = draw(lists)
    index = draw(st.integers(min_value=-len(values), max_value=len(values)))
    return values, index


@st.composite
def set_and_present_value(draw):
    values = draw(lists)
    index = draw(st.integers(min_value=-len(values), max_value=len(values) - 1))
    return set(values), values[index]


@st.composite
def set_and_missing_value(draw):
    values = draw(sets)
    missing = draw(ints.filter(lambda x: x not in values))
    return values, missing


def test_var_array_insertion():
    def fn():
        values = Array(2, 4, 6, 8)
        va = VarArray[int, 4].new()
        for i, v in enumerate(values):
            va.append(v)
            assert_true(va[i] == v)
            assert_true(len(va) == i + 1)
        return va

    assert list(run_and_validate(fn)) == [2, 4, 6, 8]


def test_var_array_setitem():
    def fn():
        va = VarArray[int, 4].new()
        va.extend(Array(2, 4, 6, 8))
        va[1] = 10
        return va

    assert list(run_and_validate(fn)) == [2, 10, 6, 8]


def test_var_array_negative_indexing():
    def fn():
        va = VarArray[int, 5].new()
        va.extend(Array(10, 20, 30, 40, 50))

        return Array(
            va[-1],
            va[-2],
            va[-3],
            va[-4],
            va[-5],
        )

    assert list(run_and_validate(fn)) == [50, 40, 30, 20, 10]


def test_var_array_negative_setitem():
    def fn():
        va = VarArray[int, 5].new()
        va.extend(Array(10, 20, 30, 40, 50))

        va[-1] = 99
        va[-2] = 88
        va[-3] = 77

        return va

    assert list(run_and_validate(fn)) == [10, 20, 77, 88, 99]


def test_var_array_negative_pop():
    def fn():
        va = VarArray[int, 5].new()
        va.extend(Array(10, 20, 30, 40, 50))

        pop1 = va.pop(-1)
        pop2 = va.pop(-2)
        length = len(va)

        results = VarArray[int, 6].new()
        results.append(pop1)
        results.append(pop2)
        results.append(length)

        for item in va:
            results.append(item)

        return results

    p1, p2, new_len, *rem = run_and_validate(fn)
    assert p1 == 50
    assert p2 == 30
    assert new_len == 3
    assert list(rem) == [10, 20, 40]


def test_var_array_del():
    def fn():
        va = VarArray[int, 4].new()
        va.extend(Array(2, 4, 6, 8))
        del va[1]
        return va

    assert list(run_and_validate(fn)) == [2, 6, 8]


def test_var_array_negative_del():
    def fn():
        va = VarArray[int, 4].new()
        va.extend(Array(2, 4, 6, 8))
        del va[-1]
        del va[-2]
        return va

    assert list(run_and_validate(fn)) == [2, 6]


def test_var_array_iadd():
    def fn():
        va = VarArray[int, 6].new()
        va.extend(Array(2, 4, 6, 8))
        va += Array(10, 12)
        return va

    assert list(run_and_validate(fn)) == [2, 4, 6, 8, 10, 12]


def test_var_array_pop():
    def fn():
        va = VarArray[int, 4].new()
        va.append(2)
        va.append(4)
        va.append(6)
        va.append(8)
        assert_true(len(va) == 4)
        assert_true(va.pop() == 8)
        assert_true(len(va) == 3)
        assert_true(va.pop(1) == 4)
        return va

    assert list(run_and_validate(fn)) == [2, 6]


def test_var_array_insert():
    def fn():
        va = VarArray[int, 4].new()
        va.append(2)
        va.append(6)
        va.insert(1, 4)
        va.insert(3, 8)
        return va

    assert list(run_and_validate(fn)) == [2, 4, 6, 8]


def test_var_array_extend():
    def fn():
        va = VarArray[int, 4].new()
        va.extend(Array(2, 4, 6, 8))
        return va

    assert list(run_and_validate(fn)) == [2, 4, 6, 8]


@given(list_and_index())
def test_var_array_pop_insert_round_trip(args):
    values_list, index = args

    values = Array(*values_list)
    value_count = len(values_list)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        assert_true(len(va) == value_count)
        expected_pop_result = values[index]
        pop_result = va.pop(index)
        assert_true(pop_result == expected_pop_result)
        assert_true(len(va) == value_count - 1)
        if index >= 0:
            insert_index = index
        elif index == -1:
            insert_index = value_count - 1
        else:
            insert_index = index + 1
        va.insert(insert_index, pop_result)
        assert_true(len(va) == value_count)
        return va

    assert list(run_and_validate(fn)) == values_list


@given(list_and_insert_index())
def test_var_array_insert_pop_round_trip(args):
    values_list, index = args

    values = Array(*values_list)
    value_count = len(values_list)

    def fn():
        va = VarArray[int, value_count + 1].new()
        va.extend(values)
        assert_true(len(va) == value_count)
        insert_value = 100
        va.insert(index, insert_value)
        assert_true(len(va) == value_count + 1)
        assert_true(va[index if index >= 0 else index - 1] == insert_value)
        pop_result = va.pop(index if index >= 0 else index - 1)
        assert_true(len(va) == value_count)
        assert_true(pop_result == insert_value)
        return va

    assert list(run_and_validate(fn)) == values_list


@given(list_and_index())
def test_remove_present_value(args):
    values_list, index = args

    values = Array(*values_list)
    value_count = len(values_list)
    to_remove = values[index]
    expected = list(values)
    expected.remove(to_remove)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        assert_true(va.remove(to_remove))
        return va

    assert list(run_and_validate(fn)) == expected


@given(lists)
def test_remove_missing_value(values_list):
    values = Array(*values_list)
    value_count = len(values_list)
    to_remove = max(values) + 1
    expected = list(values)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        assert_true(not va.remove(to_remove))
        return va

    assert list(run_and_validate(fn)) == expected


@given(list_and_index(), st.booleans())
def test_array_with_possible_uninitialized_access(args, run):
    values_list, index = args
    values = Array(*values_list)
    run = Array(run)  # We use an array so the compiler doesn't know the value
    value_count = len(values_list)

    def fn():
        va = VarArray[int, value_count].new()
        if run[0]:
            va.extend(values)
        debug_log(0)
        if run[0]:
            return va[index]
        else:
            return -1

    result = run_and_validate(fn)
    if run[0]:
        assert result == values_list[index]
    else:
        assert result == -1


def test_clear():
    def fn():
        va = VarArray[int, 4].new()
        va.extend(Array(2, 4, 6, 8))
        va.clear()
        return va

    assert list(run_and_validate(fn)) == []


@given(set_and_present_value())
def test_set_add_present(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count + 1].new()
        va.extend(values)
        assert_true(not va.set_add(value))
        return va

    # Use sorted to check that there are no duplicates
    assert sorted(run_and_validate(fn)) == sorted(values)


@given(set_and_missing_value())
def test_set_add_missing(args):
    value_set, missing = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count + 1].new()
        va.extend(values)
        assert_true(va.set_add(missing))
        return va

    # Use sorted to check that there are no duplicates
    assert sorted(run_and_validate(fn)) == sorted([*list(values), missing])


def test_set_add_full():
    def fn():
        va = VarArray[int, 4].new()
        va.extend(Array(2, 4, 6, 8))
        assert_true(not va.set_add(2))
        assert_true(not va.set_add(10))
        return va

    assert list(run_and_validate(fn)) == [2, 4, 6, 8]


@given(set_and_present_value())
def test_set_remove_present(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        assert_true(va.remove(value))
        return va

    assert sorted(run_and_validate(fn)) == sorted(value_set - {value})


@given(set_and_missing_value())
def test_set_remove_missing(args):
    value_set, missing = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        assert_true(not va.remove(missing))
        return va

    assert sorted(run_and_validate(fn)) == sorted(value_set)


@given(set_and_missing_value())
def test_set_add_remove_round_trip(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count + 1].new()
        va.extend(values)
        assert_true(va.set_add(value))
        assert_true(va.remove(value))
        return va

    assert sorted(run_and_validate(fn)) == sorted(value_set)


@given(set_and_present_value())
def test_set_remove_add_round_trip(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        va = VarArray[int, value_count + 1].new()
        va.extend(values)
        assert_true(va.remove(value))
        assert_true(not va.remove(value))
        assert_true(va.set_add(value))
        return va

    assert sorted(run_and_validate(fn)) == sorted(value_set)


@given(list_with_duplicates())
def test_var_array_count(args):
    values_list, target = args
    values = Array(*values_list)
    value_count = len(values_list)
    expected_count = values_list.count(target)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        count_result = va.count(target)
        assert_true(count_result == expected_count)
        return count_result

    assert run_and_validate(fn) == expected_count


@given(list_with_duplicates())
def test_var_array_count_present(args):
    values_list, target = args
    values = Array(*values_list)
    value_count = len(values_list)
    expected_count = values_list.count(target)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        count_result = va.count(target)
        assert_true(count_result == expected_count)
        return count_result

    assert run_and_validate(fn) == expected_count


@given(list_and_maybe_missing_value())
def test_var_array_count_maybe_missing(args):
    values_list, target = args
    values = Array(*values_list)
    value_count = len(values_list)
    expected_count = values_list.count(target)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        count_result = va.count(target)
        assert_true(count_result == expected_count)
        return count_result

    assert run_and_validate(fn) == expected_count


def test_zip_empty():
    def fn():
        va_1 = VarArray[int, 0].new()
        va_2 = VarArray[int, 0].new()
        results = VarArray[int, 0].new()
        for v1, v2 in zip(va_1, va_2):
            results.append(v1)
            results.append(v2)
        return results

    assert list(run_and_validate(fn)) == []


@given(
    lists,
)
def test_zip_single(values_list_1):
    values_1 = Array(*values_list_1)
    value_count_1 = len(values_list_1)
    result_count = value_count_1

    def fn():
        va_1 = VarArray[int, value_count_1].new()
        va_1.extend(values_1)
        results = VarArray[int, result_count].new()
        for (v1,) in zip(va_1):
            results.append(v1)
        return results

    assert list(run_and_validate(fn)) == values_list_1


@given(
    lists,
    lists,
)
def test_zip_two(values_list_1, values_list_2):
    values_1 = Array(*values_list_1)
    values_2 = Array(*values_list_2)
    value_count_1 = len(values_list_1)
    value_count_2 = len(values_list_2)
    result_count = min(value_count_1, value_count_2) * 2

    def fn():
        va_1 = VarArray[int, value_count_1].new()
        va_1.extend(values_1)
        va_2 = VarArray[int, value_count_2].new()
        va_2.extend(values_2)
        results = VarArray[int, result_count].new()
        for v1, v2 in zip(va_1, va_2):
            results.append(v1)
            results.append(v2)
        return results

    assert list(run_and_validate(fn)) == [e for pair in zip(values_list_1, values_list_2) for e in pair]


@given(
    lists,
    lists,
    lists,
)
def test_zip_three(values_list_1, values_list_2, values_list_3):
    values_1 = Array(*values_list_1)
    values_2 = Array(*values_list_2)
    values_3 = Array(*values_list_3)
    value_count_1 = len(values_list_1)
    value_count_2 = len(values_list_2)
    value_count_3 = len(values_list_3)
    result_count = min(value_count_1, value_count_2, value_count_3) * 3

    def fn():
        va_1 = VarArray[int, value_count_1].new()
        va_1.extend(values_1)
        va_2 = VarArray[int, value_count_2].new()
        va_2.extend(values_2)
        va_3 = VarArray[int, value_count_3].new()
        va_3.extend(values_3)
        results = VarArray[int, result_count].new()
        for v1, v2, v3 in zip(va_1, va_2, va_3):
            results.append(v1)
            results.append(v2)
            results.append(v3)
        return results

    assert list(run_and_validate(fn)) == [
        e for triple in zip(values_list_1, values_list_2, values_list_3) for e in triple
    ]


def test_map_single():
    def fn():
        va = VarArray[int, 4].new()
        va.extend(Array(1, 2, 3, 4))
        results = VarArray[int, 4].new()
        for x in map(lambda x: x * 2, va):
            results.append(x)
        return results

    assert list(run_and_validate(fn)) == [2, 4, 6, 8]


def test_map_two_args():
    def fn():
        va1 = VarArray[int, 4].new()
        va2 = VarArray[int, 4].new()
        va1.extend(Array(1, 2, 3, 4))
        va2.extend(Array(10, 20, 30, 40))
        results = VarArray[int, 4].new()
        for x in map(lambda x, y: x + y, va1, va2):
            results.append(x)
        return results

    assert list(run_and_validate(fn)) == [11, 22, 33, 44]


def test_map_three_args():
    def fn():
        va1 = VarArray[int, 3].new()
        va2 = VarArray[int, 3].new()
        va3 = VarArray[int, 3].new()
        va1.extend(Array(1, 2, 3))
        va2.extend(Array(10, 20, 30))
        va3.extend(Array(100, 200, 300))
        results = VarArray[int, 3].new()
        for x in map(lambda x, y, z: x + y + z, va1, va2, va3):
            results.append(x)
        return results

    assert list(run_and_validate(fn)) == [111, 222, 333]


def test_filter_basic():
    def fn():
        va = VarArray[int, 6].new()
        va.extend(Array(1, 2, 3, 4, 5, 6))
        results = VarArray[int, 6].new()
        for x in filter(lambda x: x % 2 == 0, va):
            results.append(x)
        return results

    assert list(run_and_validate(fn)) == [2, 4, 6]


@given(lists)
def test_filter_property(values_list):
    values = Array(*values_list)
    value_count = len(values_list)

    def fn():
        va = VarArray[int, value_count].new()
        va.extend(values)
        results = VarArray[int, value_count].new()
        for x in filter(lambda x: x > 0, va):
            results.append(x)
        return results

    assert list(run_and_validate(fn)) == [x for x in values_list if x > 0]


def test_map_filter_combination():
    def fn():
        va = VarArray[int, 6].new()
        va.extend(Array(1, 2, 3, 4, 5, 6))
        results = VarArray[int, 6].new()
        for x in map(lambda x: x * 2, filter(lambda x: x % 2 == 0, va)):
            results.append(x)
        return results

    assert list(run_and_validate(fn)) == [4, 8, 12]


def test_filter_map_combination():
    def fn():
        va = VarArray[int, 6].new()
        va.extend(Array(1, 2, 3, 4, 5, 6))
        results = VarArray[int, 6].new()
        for x in filter(lambda x: x > 5, map(lambda x: x * 2, va)):
            results.append(x)
        return results

    assert list(run_and_validate(fn)) == [6, 8, 10, 12]


def test_any_true():
    def fn():
        va = VarArray[bool, 6].new()
        va.extend(Array(True, False, True, False, True, False))
        return any(va)

    assert run_and_validate(fn)


def test_any_false():
    def fn():
        va = VarArray[bool, 6].new()
        va.extend(Array(False, False, False, False, False, False))
        return any(va)

    assert not run_and_validate(fn)


def test_any_empty():
    def fn():
        va = VarArray[bool, 0].new()
        return any(va)

    assert not run_and_validate(fn)


def test_all_true():
    def fn():
        va = VarArray[bool, 6].new()
        va.extend(Array(True, True, True, True, True, True))
        return all(va)

    assert run_and_validate(fn)


def test_all_false():
    def fn():
        va = VarArray[bool, 6].new()
        va.extend(Array(True, False, True, False, True, False))
        return all(va)

    assert not run_and_validate(fn)


def test_all_empty():
    def fn():
        va = VarArray[bool, 0].new()
        return all(va)

    assert run_and_validate(fn)


def test_var_array_truthiness_empty():
    def fn():
        x = VarArray[int, 1].new()
        return 1 if x else 0

    assert run_and_validate(fn) == 0


def test_var_array_truthiness_non_empty():
    def fn():
        x = VarArray[int, 3].new()
        x.append(1)
        return 1 if x else 0

    assert run_and_validate(fn) == 1



from __future__ import annotations

from collections.abc import Callable
from typing import Any

from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.transient import TransientValue
from sonolus.script.internal.value import Value
from sonolus.script.num import Num
from sonolus.script.values import copy, zeros


class Maybe[T](TransientValue):
    """A type that either has a value or is empty.

    `Maybe` has special behavior when returned from a function: unlike records and arrays, it may be returned from
    multiple places in a function, provided that all but one return statement returns the literal
    [`Nothing`][sonolus.script.maybe.Nothing].

    Storing values of this type in a Record, Array, or Archetype is not supported.

    Usage:
        ```python
        def fn(a, b):
            if a:
                return Some(b)
            else:
                return Nothing

        result = fn(..., ...)
        if result.is_some:
            value = result.get()
            ...
        ```
    """

    _present: Num
    _value: T

    def __init__(self, *, present: bool, value: T):
        self._present = Num._accept_(present)
        self._value = validate_value(value)

    @property
    @meta_fn
    def is_some(self) -> bool:
        """Check if the value is present."""
        if ctx():
            if self._present._is_py_():
                # Makes this a compile time constant.
                return self._present
            return self._present._get_readonly_()
        else:
            return self._present._as_py_()

    @property
    def is_nothing(self) -> bool:
        """Check if the value is empty."""
        return not self.is_some

    def get(self) -> T:
        """Get the value if present, otherwise raise an error."""
        assert self.is_some
        return self.get_unsafe()

    @meta_fn
    def get_unsafe(self) -> T:
        if ctx():
            return self._value
        else:
            return self._value._as_py_()

    def map[R](self, fn: Callable[[T], R], /) -> Maybe[R]:
        """Map the contained value to a new value using the provided function.

        If the value is not present, returns [`Nothing`][sonolus.script.maybe.Nothing].

        Args:
            fn: A function that takes the contained value and returns a new value.

        Returns:
            A [`Maybe`][sonolus.script.maybe.Maybe] instance containing the result of the function if the value
            is present, otherwise [`Nothing`][sonolus.script.maybe.Nothing].
        """
        if self.is_some:
            return Some(fn(self.get_unsafe()))
        return Nothing

    def flat_map[R](self, fn: Callable[[T], Maybe[R]], /) -> Maybe[R]:
        """Flat map the contained value to a new [`Maybe`][sonolus.script.maybe.Maybe] using the provided function.

        If the value is not present, returns [`Nothing`][sonolus.script.maybe.Nothing].

        Args:
            fn: A function that takes the contained value and returns a new [`Maybe`][sonolus.script.maybe.Maybe].

        Returns:
            A [`Maybe`][sonolus.script.maybe.Maybe] instance containing the result of the function if the value
            is present, otherwise [`Nothing`][sonolus.script.maybe.Nothing].
        """
        if self.is_some:
            return fn(self.get_unsafe())
        return Nothing

    def or_default(self, default: T) -> T:
        """Return a copy of the contained value if present, otherwise return a copy of the given default value.

        Args:
            default: The default value to return if the contained value is not present.

        Returns:
            A copy of the contained value if present, otherwise a copy of the default value.
        """
        result = _box(copy(default))
        if self.is_some:
            result.value = self.get_unsafe()
        return result.value

    @meta_fn
    def or_else(self, fn: Callable[[], T], /) -> T:
        """Return a copy of the contained value if present, otherwise return a copy of the result of the given function.

        Args:
            fn: A function that returns a value to use if the contained value is not present.

        Returns:
            A copy of the contained value if present, otherwise a copy of the result of calling the function.
        """
        from sonolus.backend.visitor import compile_and_call

        if ctx():
            if self.is_some._is_py_():  # type: ignore
                if self.is_some._as_py_():  # type: ignore
                    return copy(self.get_unsafe())
                else:
                    return copy(compile_and_call(fn))
            else:
                return compile_and_call(self._or_else, fn)
        elif self.is_some:
            return copy(self.get_unsafe())
        else:
            return copy(fn())

    def _or_else(self, fn: Callable[[], T], /) -> T:
        result = _box(zeros(self.contained_type))
        if self.is_some:
            result.value = self.get_unsafe()
        else:
            result.value = fn()
        return result.value

    @property
    def tuple(self) -> tuple[bool, T]:
        """Return whether the value is present and a copy of the contained value if present as a tuple.

        If the value is not present, the tuple will contain `False` and a zero initialized value of the contained type.
        """
        result_value = _box(zeros(self.contained_type))
        if self.is_some:
            result_value.value = self.get_unsafe()
        return self.is_some, result_value.value

    @property
    @meta_fn
    def contained_type(self):
        return type(self._value)

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        return isinstance(value, cls)

    @classmethod
    def _accept_(cls, value: Any) -> Maybe[T]:
        if not cls._accepts_(value):
            raise TypeError(f"Cannot accept value of type {type(value).__name__} as {cls.__name__}.")
        return value

    def _is_py_(self) -> bool:
        return not self._present or not isinstance(self._value, Value) or self._value._is_py_()

    def _as_py_(self) -> Any:
        if not self._is_py_():
            raise ValueError("Not a python value")
        return self

    def _copy_from_(self, value: Any):
        raise TypeError("Maybe does not support mutation.")

    def _copy_(self) -> Maybe[T]:
        raise TypeError("Maybe does not support copying.")

    def _set_(self, value: Any):
        if not self._accepts_(value):
            raise TypeError(f"Cannot set value of type {type(value).__name__} to {self.__class__.__name__}.")
        if value is not Nothing and self._value is not value._value:
            raise TypeError(f"Cannot set value of type {type(value._value).__name__} to {self.__class__.__name__}.")
        self._present._set_(value._present)

    @classmethod
    def _get_merge_target_(cls, values: list[Any]) -> Any:
        if not all(isinstance(v, cls) for v in values):
            return NotImplemented
        distinct = []
        seen_ids = set()
        for v in values:
            if v is Nothing:
                continue
            if id(v._value) not in seen_ids:
                distinct.append(v)
                seen_ids.add(id(v._value))
        match distinct:
            case []:
                return Nothing
            case [v]:
                return Maybe(present=Num._alloc_(), value=v._value)
            case _:
                return NotImplemented


def Some[T](value: T) -> Maybe[T]:  # noqa: N802
    """Create a [`Maybe`][sonolus.script.maybe.Maybe] instance with a value.

    Args:
        value: The contained value.

    Returns:
        A [`Maybe`][sonolus.script.maybe.Maybe] instance that contains the provided value.
    """
    return Maybe(present=True, value=value)


Nothing: Maybe[Any] = Maybe(present=False, value=None)  # type: ignore

# Note: has to come after the definition to hide the definition in the docs.
Nothing: Maybe[Any]
"""The empty [`Maybe`][sonolus.script.maybe.Maybe] instance."""


@meta_fn
def _box(value):
    from sonolus.script.containers import Box

    return Box(value)



"""Tests for assert statements.

PYTEST_DONT_REWRITE
"""

from sonolus.script.debug import assert_false, assert_true
from tests.script.conftest import run_compiled


def test_assertion_succeeds():
    def fn():
        assert True, "Message"
        return 1

    assert run_compiled(fn) == 1


def test_assertion_fails():
    def fn():
        assert False, "Message"  # noqa: B011, PT015
        # noinspection PyUnreachableCode
        return 1

    assert run_compiled(fn) == 0


def test_assert_true_succeeds():
    def fn():
        assert_true(True)
        return 1

    assert run_compiled(fn) == 1


def test_assert_true_fails():
    def fn():
        assert_true(False)
        return 1

    assert run_compiled(fn) == 0


def test_assert_false_succeeds():
    def fn():
        assert_false(False)
        return 1

    assert run_compiled(fn) == 1


def test_assert_false_fails():
    def fn():
        assert_false(True)
        return 1

    assert run_compiled(fn) == 0



import argparse
import contextlib
import http.server
import importlib
import json
import shutil
import socket
import socketserver
import sys
from pathlib import Path
from time import perf_counter

from sonolus.backend.excepthook import print_simple_traceback
from sonolus.backend.optimize.optimize import FAST_PASSES, MINIMAL_PASSES, STANDARD_PASSES
from sonolus.build.engine import no_gil, package_engine, validate_engine
from sonolus.build.level import package_level_data
from sonolus.build.project import build_project_to_collection, get_project_schema
from sonolus.script.internal.error import CompilationError
from sonolus.script.project import BuildConfig, Project


def find_default_module() -> str | None:
    current_dir = Path.cwd()

    potential_modules = []

    project_files = list(current_dir.glob("*/project.py"))
    potential_modules.extend(str(f.parent.relative_to(current_dir)).replace("/", ".") for f in project_files)

    init_files = list(current_dir.glob("*/__init__.py"))
    potential_modules.extend(str(f.parent.relative_to(current_dir)).replace("/", ".") for f in init_files)

    potential_modules = [m for m in set(potential_modules) if m]

    return potential_modules[0] if len(potential_modules) == 1 else None


def import_project(module_path: str) -> Project | None:
    try:
        current_dir = Path.cwd()
        if current_dir not in sys.path:
            sys.path.insert(0, str(current_dir))

        project = None

        try:
            module = importlib.import_module(module_path)
            project = getattr(module, "project", None)
        except ImportError as e:
            if not str(e).endswith(f"'{module_path}'"):
                # It's an error from the module itself
                raise

        if project is None:
            try:
                project_module = importlib.import_module(f"{module_path}.project")
                project = getattr(project_module, "project", None)
            except ImportError as e:
                if not str(e).endswith(f"'{module_path}.project'"):
                    raise

        if project is None:
            print(f"Error: No Project instance found in module {module_path} or {module_path}.project")
            return None

        return project
    except Exception as e:
        print(f"Error: Failed to import project: {e}")
        raise e from None


def build_project(project: Project, build_dir: Path, config: BuildConfig):
    dist_dir = build_dir / "dist"
    levels_dir = dist_dir / "levels"
    shutil.rmtree(dist_dir, ignore_errors=True)
    dist_dir.mkdir(parents=True, exist_ok=True)
    levels_dir.mkdir(parents=True, exist_ok=True)

    package_engine(project.engine.data, config).write(dist_dir / "engine")

    for level in project.levels:
        level_path = levels_dir / level.name
        level_path.write_bytes(package_level_data(level.data))


def validate_project(project: Project, config: BuildConfig):
    validate_engine(project.engine.data, config)


def build_collection(project: Project, build_dir: Path, config: BuildConfig | None):
    site_dir = build_dir / "site"
    shutil.rmtree(site_dir, ignore_errors=True)
    site_dir.mkdir(parents=True, exist_ok=True)

    collection = build_project_to_collection(project, config)
    collection.write(site_dir)


def get_local_ips():
    hostname = socket.gethostname()
    local_ips = []

    with contextlib.suppress(socket.gaierror):
        local_ips.append(socket.gethostbyname(socket.getfqdn()))

    try:
        for info in socket.getaddrinfo(hostname, None):
            ip = info[4][0]
            if not ip.startswith("127.") and ":" not in ip:
                local_ips.append(ip)
    except socket.gaierror:
        pass

    return sorted(set(local_ips))


def run_server(base_dir: Path, port: int = 8000):
    class DirectoryHandler(http.server.SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, directory=str(base_dir), **kwargs)

    with socketserver.TCPServer(("", port), DirectoryHandler) as httpd:
        local_ips = get_local_ips()
        print(f"Server started on port {port}")
        print("Available on:")
        for ip in local_ips:
            print(f"  http://{ip}:{port}")
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            print("\nStopping server...")
            httpd.shutdown()


def get_config(args: argparse.Namespace) -> BuildConfig:
    if hasattr(args, "optimize_minimal") and args.optimize_minimal:
        optimization_passes = MINIMAL_PASSES
    elif hasattr(args, "optimize_fast") and args.optimize_fast:
        optimization_passes = FAST_PASSES
    elif hasattr(args, "optimize_standard") and args.optimize_standard:
        optimization_passes = STANDARD_PASSES
    else:
        optimization_passes = FAST_PASSES if args.command == "dev" else STANDARD_PASSES

    if any(hasattr(args, attr) and getattr(args, attr) for attr in ["play", "watch", "preview", "tutorial"]):
        build_play = hasattr(args, "play") and args.play
        build_watch = hasattr(args, "watch") and args.watch
        build_preview = hasattr(args, "preview") and args.preview
        build_tutorial = hasattr(args, "tutorial") and args.tutorial
    else:
        build_play = build_watch = build_preview = build_tutorial = True

    return BuildConfig(
        passes=optimization_passes,
        build_play=build_play,
        build_watch=build_watch,
        build_preview=build_preview,
        build_tutorial=build_tutorial,
    )


def main():
    parser = argparse.ArgumentParser(description="Sonolus project build and development tools")
    subparsers = parser.add_subparsers(dest="command", required=True)

    def add_common_arguments(parser):
        optimization_group = parser.add_mutually_exclusive_group()
        optimization_group.add_argument(
            "-O0", "--optimize-minimal", action="store_true", help="Use minimal optimization passes"
        )
        optimization_group.add_argument(
            "-O1", "--optimize-fast", action="store_true", help="Use fast optimization passes"
        )
        optimization_group.add_argument(
            "-O2", "--optimize-standard", action="store_true", help="Use standard optimization passes"
        )

        build_components = parser.add_argument_group("build components")
        build_components.add_argument("--play", action="store_true", help="Build play component")
        build_components.add_argument("--watch", action="store_true", help="Build watch component")
        build_components.add_argument("--preview", action="store_true", help="Build preview component")
        build_components.add_argument("--tutorial", action="store_true", help="Build tutorial component")

    build_parser = subparsers.add_parser("build")
    build_parser.add_argument(
        "module",
        type=str,
        nargs="?",
        help="Module path (e.g., 'module.name'). If omitted, will auto-detect if only one module exists.",
    )
    build_parser.add_argument("--build-dir", type=str, default="./build")
    add_common_arguments(build_parser)

    dev_parser = subparsers.add_parser("dev")
    dev_parser.add_argument(
        "module",
        type=str,
        nargs="?",
        help="Module path (e.g., 'module.name'). If omitted, will auto-detect if only one module exists.",
    )
    dev_parser.add_argument("--build-dir", type=str, default="./build")
    dev_parser.add_argument("--port", type=int, default=8000)
    add_common_arguments(dev_parser)

    schema_parser = subparsers.add_parser("schema")
    schema_parser.add_argument(
        "module",
        type=str,
        nargs="?",
        help="Module path (e.g., 'module.name'). If omitted, will auto-detect if only one module exists.",
    )

    check_parser = subparsers.add_parser("check")
    check_parser.add_argument(
        "module",
        type=str,
        nargs="?",
        help="Module path (e.g., 'module.name'). If omitted, will auto-detect if only one module exists.",
    )
    add_common_arguments(check_parser)

    args = parser.parse_args()

    if not args.module:
        default_module = find_default_module()
        if default_module:
            if args.command != "schema":
                print(f"Using auto-detected module: {default_module}")
            args.module = default_module
        else:
            parser.error("Module argument is required when multiple or no modules are found")

    if no_gil():
        print("Multithreading is enabled")
    if hasattr(sys, "_jit") and sys._jit.is_enabled():
        print("Python JIT is enabled")

    start_time = perf_counter()
    project = import_project(args.module)
    end_time = perf_counter()
    if project is None:
        sys.exit(1)
    print(f"Project imported in {end_time - start_time:.2f}s")

    try:
        if args.command == "build":
            build_dir = Path(args.build_dir)
            start_time = perf_counter()
            config = get_config(args)
            build_project(project, build_dir, config)
            end_time = perf_counter()
            print(f"Project built successfully to '{build_dir.resolve()}' in {end_time - start_time:.2f}s")
        elif args.command == "dev":
            build_dir = Path(args.build_dir)
            start_time = perf_counter()
            config = get_config(args)
            build_collection(project, build_dir, config)
            end_time = perf_counter()
            print(f"Build finished in {end_time - start_time:.2f}s")
            run_server(build_dir / "site", port=args.port)
        elif args.command == "schema":
            print(json.dumps(get_project_schema(project), indent=2))
        elif args.command == "check":
            start_time = perf_counter()
            config = get_config(args)
            validate_project(project, config)
            end_time = perf_counter()
            print(f"Project validation completed successfully in {end_time - start_time:.2f}s")
    except CompilationError:
        exc_info = sys.exc_info()
        print_simple_traceback(*exc_info)



from collections.abc import Callable, Sequence
from concurrent.futures import Executor, Future

from sonolus.backend.finalize import cfg_to_engine_node
from sonolus.backend.ir import IRConst, IRInstr
from sonolus.backend.mode import Mode
from sonolus.backend.ops import Op
from sonolus.backend.optimize.flow import BasicBlock
from sonolus.backend.optimize.optimize import STANDARD_PASSES
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig, run_passes
from sonolus.backend.visitor import compile_and_call_at_definition
from sonolus.build.node import OutputNodeGenerator
from sonolus.script.archetype import _BaseArchetype
from sonolus.script.internal.callbacks import CallbackInfo
from sonolus.script.internal.context import (
    CallbackContextState,
    Context,
    GlobalContextState,
    ReadOnlyMemory,
    context_to_cfg,
    ctx,
    using_ctx,
)
from sonolus.script.internal.error import CompilationError
from sonolus.script.num import _is_num


def compile_mode(
    mode: Mode,
    rom: ReadOnlyMemory,
    archetypes: list[type[_BaseArchetype]] | None,
    global_callbacks: list[tuple[CallbackInfo, Callable]] | None,
    passes: Sequence[CompilerPass] | None = None,
    thread_pool: Executor | None = None,
    validate_only: bool = False,
) -> dict:
    if passes is None:
        passes = STANDARD_PASSES

    global_state = GlobalContextState(
        mode,
        {a: i for i, a in enumerate(archetypes)} if archetypes is not None else None,
        rom,
    )
    nodes = OutputNodeGenerator()
    results = {}

    def process_callback(
        cb_info: CallbackInfo,
        cb: Callable,
        arch: type[_BaseArchetype] | None = None,
    ) -> tuple[str, int] | tuple[str, dict]:
        """Compile a single callback to a node.

        Returns either:
        - (cb_info.name, node_index) for global callbacks, or
        - (cb_info.name, {"index": node_index, "order": cb_order}) for archetype callbacks.
        """
        cfg = callback_to_cfg(global_state, cb, cb_info.name, arch)
        if validate_only:
            if arch is not None:
                cb_order = getattr(cb, "_callback_order_", 0)
                return cb_info.name, {"index": 0, "order": cb_order}
            else:
                return cb_info.name, 0
        cfg = run_passes(cfg, passes, OptimizerConfig(mode=mode, callback=cb_info.name))
        node = cfg_to_engine_node(cfg)
        node_index = nodes.add(node)

        if arch is not None:
            cb_order = getattr(cb, "_callback_order_", 0)
            return cb_info.name, {"index": node_index, "order": cb_order}
        else:
            return cb_info.name, node_index

    all_futures = {}
    base_archetype_entries = {}

    if archetypes is not None:
        base_archetypes = {getattr(a, "_derived_base_", a) for a in archetypes}

        for archetype in base_archetypes:
            archetype._init_fields()

            archetype_data = {
                "name": archetype.name,
                "hasInput": archetype.is_scored,
                "imports": [{"name": name, "index": index} for name, index in archetype._imported_keys_.items()],
            }
            if mode == Mode.PLAY:
                archetype_data["exports"] = [*archetype._exported_keys_]

            callback_items = [
                (cb_name, cb_info, getattr(archetype, cb_name))
                for cb_name, cb_info in archetype._supported_callbacks_.items()
                if getattr(archetype, cb_name) not in archetype._default_callbacks_
            ]

            if thread_pool is not None:
                for cb_name, cb_info, cb in callback_items:
                    cb_order = getattr(cb, "_callback_order_", 0)
                    if not cb_info.supports_order and cb_order != 0:
                        raise ValueError(f"Callback '{cb_name}' does not support a non-zero order")
                    f: Future = thread_pool.submit(process_callback, cb_info, cb, archetype)
                    all_futures[f] = ("archetype", archetype_data, cb_name)
            else:
                for cb_name, cb_info, cb in callback_items:
                    cb_order = getattr(cb, "_callback_order_", 0)
                    if not cb_info.supports_order and cb_order != 0:
                        raise ValueError(f"Callback '{cb_name}' does not support a non-zero order")
                    cb_name, result_data = process_callback(cb_info, cb, archetype)
                    archetype_data[cb_name] = result_data

            base_archetype_entries[archetype] = archetype_data

    if global_callbacks is not None and thread_pool is not None:
        for cb_info, cb in global_callbacks:
            f: Future = thread_pool.submit(process_callback, cb_info, cb, None)
            all_futures[f] = ("global", None, cb_info.name)

    if thread_pool is not None:
        for f, (callback_type, archetype_data, cb_name) in all_futures.items():
            cb_name, result_data = f.result()
            if callback_type == "archetype":
                archetype_data[cb_name] = result_data
            else:  # global callback
                results[cb_name] = result_data
    elif global_callbacks is not None:
        for cb_info, cb in global_callbacks:
            cb_name, node_index = process_callback(cb_info, cb, None)
            results[cb_name] = node_index

    if archetypes is not None:
        results["archetypes"] = [
            {**base_archetype_entries[getattr(a, "_derived_base_", a)], "name": a.name, "hasInput": a.is_scored}
            for a in archetypes
        ]

    results["nodes"] = nodes.get()
    return results


def callback_to_cfg(
    global_state: GlobalContextState,
    callback: Callable,
    name: str,
    archetype: type[_BaseArchetype] | None = None,
) -> BasicBlock:
    try:
        # Default to no_eval=True for performance unless there's an error.
        return _callback_to_cfg(global_state, callback, name, archetype, no_eval=True)
    except CompilationError:
        return _callback_to_cfg(global_state, callback, name, archetype, no_eval=False)


def _callback_to_cfg(
    global_state: GlobalContextState,
    callback: Callable,
    name: str,
    archetype: type[_BaseArchetype] | None,
    no_eval: bool,
) -> BasicBlock:
    callback_state = CallbackContextState(name, no_eval=no_eval)
    context = Context(global_state, callback_state)
    with using_ctx(context):
        if archetype is not None:
            result = compile_and_call_at_definition(callback, archetype._for_compilation())
        else:
            result = compile_and_call_at_definition(callback)
        if _is_num(result):
            ctx().add_statements(IRInstr(Op.Break, [IRConst(1), result.ir()]))
    return context_to_cfg(context)



from __future__ import annotations

from typing import cast

from pydori.lib.layer import LAYER_ARROW, LAYER_NOTE, get_z
from pydori.lib.note import NoteKind, get_note_arrow_sprite, get_note_body_sprite
from pydori.lib.options import Options
from pydori.preview.layout import (
    PreviewData,
    layout_preview_directional_flick_arrow,
    layout_preview_flick_arrow,
    layout_preview_note,
)
from sonolus.script.archetype import PreviewArchetype, StandardImport, entity_data, imported
from sonolus.script.timing import beat_to_time


class PreviewNote(PreviewArchetype):
    """Common archetype for notes."""

    lane: float = imported()
    beat: StandardImport.BEAT = imported()
    direction: int = imported()

    target_time: float = entity_data()

    def preprocess(self):
        if Options.mirror:
            self.lane = -self.lane
            self.direction = -self.direction

        self.target_time = beat_to_time(self.beat)

        PreviewData.last_time = max(PreviewData.last_time, self.target_time)
        PreviewData.last_beat = max(PreviewData.last_beat, self.beat)

    def render(self):
        self.draw_body()
        self.draw_arrow()

    def draw_body(self):
        if self.kind == NoteKind.HOLD_ANCHOR:
            return
        body_sprite = get_note_body_sprite(self.kind, self.direction)
        layout = layout_preview_note(self.lane, self.target_time)
        body_sprite.draw(layout, z=get_z(LAYER_NOTE, lane=self.lane, y=self.target_time))

    def draw_arrow(self):
        arrow_sprite = get_note_arrow_sprite(self.kind, self.direction)
        match self.kind:
            case NoteKind.FLICK:
                layout = layout_preview_flick_arrow(self.lane, self.target_time)
                arrow_sprite.draw(layout, z=get_z(LAYER_ARROW, lane=self.lane, y=self.target_time))
            case NoteKind.DIRECTIONAL_FLICK:
                for i in range(abs(self.direction)):
                    lane_offset = (i + 1) * (1 if self.direction > 0 else -1)
                    arrow_lane = self.lane + lane_offset
                    layout = layout_preview_directional_flick_arrow(
                        arrow_lane, self.target_time, direction=self.direction
                    )
                    arrow_sprite.draw(layout, z=get_z(LAYER_ARROW, lane=arrow_lane, y=self.target_time))
            case _:
                pass

    @property
    def kind(self) -> NoteKind:
        return cast(NoteKind, self.key)


PreviewTapNote = PreviewNote.derive("Tap", is_scored=True, key=NoteKind.TAP)
PreviewFlickNote = PreviewNote.derive("Flick", is_scored=True, key=NoteKind.FLICK)
PreviewDirectionalFlickNote = PreviewNote.derive("DirectionalFlick", is_scored=True, key=NoteKind.DIRECTIONAL_FLICK)
PreviewHoldHeadNote = PreviewNote.derive("HoldHead", is_scored=True, key=NoteKind.HOLD_HEAD)
PreviewHoldTickNote = PreviewNote.derive("HoldTick", is_scored=True, key=NoteKind.HOLD_TICK)
PreviewHoldAnchorNote = PreviewNote.derive("HoldAnchor", is_scored=False, key=NoteKind.HOLD_ANCHOR)
PreviewHoldEndNote = PreviewNote.derive("HoldEnd", is_scored=True, key=NoteKind.HOLD_END)

ALL_PREVIEW_NOTE_TYPES = (
    PreviewTapNote,
    PreviewFlickNote,
    PreviewDirectionalFlickNote,
    PreviewHoldHeadNote,
    PreviewHoldTickNote,
    PreviewHoldAnchorNote,
    PreviewHoldEndNote,
)



from math import isfinite, isinf, isnan

from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet
from sonolus.backend.node import ConstantNode, EngineNode, FunctionNode
from sonolus.backend.ops import Op
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_reverse_postorder
from sonolus.backend.place import BlockPlace


def cfg_to_engine_node(entry: BasicBlock):
    block_indexes = {block: i for i, block in enumerate(traverse_cfg_reverse_postorder(entry))}
    block_statements = []
    for block in block_indexes:
        statements = []
        statements.extend(ir_to_engine_node(stmt) for stmt in block.statements)
        outgoing = {
            edge.cond: edge.dst for edge in sorted(block.outgoing, key=lambda edge: (edge.cond is None, edge.cond))
        }
        match outgoing:
            case {**other} if not other:
                statements.append(ConstantNode(value=len(block_indexes)))
            case {None: target, **other} if not other:
                statements.append(ConstantNode(value=block_indexes[target]))
            case {0: f_branch, None: t_branch, **other} if not other:
                statements.append(
                    FunctionNode(
                        func=Op.If,
                        args=[
                            ir_to_engine_node(block.test),
                            ConstantNode(value=block_indexes[t_branch]),
                            ConstantNode(value=block_indexes[f_branch]),
                        ],
                    )
                )
            case {None: default_branch, **other} if len(other) == 1:
                cond, cond_branch = next(iter(other.items()))
                statements.append(
                    FunctionNode(
                        func=Op.If,
                        args=[
                            ir_to_engine_node(IRPureInstr(Op.Equal, args=[block.test, IRConst(cond)])),
                            ConstantNode(value=block_indexes[cond_branch]),
                            ConstantNode(value=block_indexes[default_branch]),
                        ],
                    )
                )
            case dict() as targets:
                args = [ir_to_engine_node(block.test)]
                default = len(block_indexes)
                conds = [cond for cond in targets if cond is not None]
                if min(conds) == 0 and max(conds) == len(conds) - 1 and all(int(cond) == cond for cond in conds):
                    args.extend(ConstantNode(value=block_indexes[targets[cond]]) for cond in range(len(conds)))
                    if None in targets:
                        default = block_indexes[targets[None]]
                    args.append(ConstantNode(value=default))
                    statements.append(FunctionNode(Op.SwitchIntegerWithDefault, args))
                else:
                    for cond, target in targets.items():
                        if cond is None:
                            default = block_indexes[target]
                            continue
                        args.append(ConstantNode(value=cond))
                        args.append(ConstantNode(value=block_indexes[target]))
                    args.append(ConstantNode(value=default))
                    statements.append(FunctionNode(Op.SwitchWithDefault, args))
        block_statements.append(FunctionNode(Op.Execute, statements))
    block_statements.append(ConstantNode(value=0))
    return FunctionNode(Op.Block, [FunctionNode(Op.JumpLoop, block_statements)])


def ir_to_engine_node(stmt) -> EngineNode:
    match stmt:
        case int(value) | float(value) | IRConst(value=int(value) | float(value)):
            value = float(value)
            if value.is_integer():
                return ConstantNode(value=int(value))
            elif isfinite(value):
                return ConstantNode(value=value)
            elif isinf(value):
                # Read values from ROM
                return FunctionNode(Op.Get, args=[ConstantNode(value=3000), ConstantNode(value=1 if value > 0 else 2)])
            elif isnan(value):
                # Read value from ROM
                return FunctionNode(Op.Get, args=[ConstantNode(value=3000), ConstantNode(value=0)])
            else:
                raise ValueError(f"Invalid constant value: {value}")
        case IRPureInstr(op=op, args=args) | IRInstr(op=op, args=args):
            return FunctionNode(func=op, args=[ir_to_engine_node(arg) for arg in args])
        case IRGet(place=place):
            return ir_to_engine_node(place)
        case BlockPlace() as place:
            if place.offset == 0:
                index = ir_to_engine_node(place.index)
            elif place.index == 0:
                index = ConstantNode(value=place.offset)
            else:
                index = FunctionNode(
                    func=Op.Add, args=[ir_to_engine_node(place.index), ConstantNode(value=place.offset)]
                )
            return FunctionNode(func=Op.Get, args=[ir_to_engine_node(place.block), index])
        case IRSet(place=place, value=value):
            return FunctionNode(func=Op.Set, args=[*ir_to_engine_node(place).args, ir_to_engine_node(value)])
        case _:
            raise TypeError(f"Unsupported IR statement: {stmt}")



import itertools
import os
import random
import sys
from collections.abc import Callable
from datetime import timedelta
from types import CellType

from hypothesis import settings

from sonolus.backend.blocks import PlayBlock
from sonolus.backend.finalize import cfg_to_engine_node
from sonolus.backend.interpret import Interpreter
from sonolus.backend.mode import Mode
from sonolus.backend.optimize.optimize import FAST_PASSES, MINIMAL_PASSES, STANDARD_PASSES
from sonolus.backend.optimize.passes import OptimizerConfig, run_passes
from sonolus.backend.place import BlockPlace
from sonolus.backend.visitor import compile_and_call
from sonolus.build.compile import callback_to_cfg
from sonolus.script.debug import debug_log_callback, simulation_context
from sonolus.script.internal.context import GlobalContextState, ctx
from sonolus.script.internal.error import CompilationError
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.tuple_impl import TupleImpl
from sonolus.script.num import Num
from sonolus.script.vec import Vec2

PRIMARY_PYTHON_VERSION = (3, 14)


def is_ci() -> bool:
    return os.getenv("CI", "false").lower() in {"true", "1"}


settings.register_profile(
    "default",
    settings.get_profile("default"),
    max_examples=100,
    deadline=timedelta(seconds=10),
)
settings.register_profile(
    "ci",
    settings.get_profile("ci"),
    max_examples=40,
    deadline=timedelta(seconds=10),
)
settings.load_profile("ci" if is_ci() else "default")

optimization_levels = [
    MINIMAL_PASSES,
    FAST_PASSES,
    STANDARD_PASSES,
]

if is_ci() and sys.version_info < PRIMARY_PYTHON_VERSION:
    optimization_levels = [STANDARD_PASSES]


def compile_fn(callback: Callable):
    global_state = GlobalContextState(Mode.PLAY)
    return callback_to_cfg(global_state, callback, ""), global_state.rom.values


def run_and_validate[**P, R](
    fn: Callable[P, R], *args: P.args, use_simulation_context: bool = False, **kwargs: P.kwargs
) -> R:
    """Runs a function as a regular function and as a compiled function, and checks that the results are the same."""
    exception = None
    regular_result = None
    log_entries = []

    def log_cb(x):
        log_entries.append(x)
        return 0

    debug_log_callback_token = debug_log_callback.set(log_cb)
    try:
        if use_simulation_context:
            with simulation_context():
                regular_result = fn(*args, **kwargs)
        else:
            regular_result = fn(*args, **kwargs)
    except Exception as e:
        exception = e
    finally:
        debug_log_callback.reset(debug_log_callback_token)

    result_type = None

    @meta_fn
    def run_compiled():
        nonlocal result_type
        result = compile_and_call(fn, *args, **kwargs)
        # If terminated, this line won't run
        # We can check whether this value is 1 to see if the compiled function finished without terminating
        Num._from_place_(BlockPlace(-1, 0))._set_(Num(1))
        result_type = type(result)
        target = result_type._from_place_(BlockPlace(-2, 0))
        if result_type._is_value_type_():
            target._set_(result)
        else:
            target._copy_from_(result)
        return result

    @meta_fn
    def run_compiled_with_closure_from_rom():
        nonlocal result_type
        closure: list[CellType] | None = getattr(fn, "__closure__", None)
        if closure:
            original_values = [cell.cell_contents for cell in closure]

            def value_to_rom(v):
                value = validate_value(v)
                if isinstance(value, TupleImpl):
                    return TupleImpl(tuple(value_to_rom(entry) for entry in value.value))
                else:
                    return type(value)._from_place_(ctx().rom[tuple(value._to_list_())])

            try:
                for cell, original_value in zip(closure, original_values, strict=True):
                    cell.cell_contents = value_to_rom(original_value)
                result = compile_and_call(fn, *args, **kwargs)
            finally:
                for cell, original_value in zip(closure, original_values, strict=True):
                    cell.cell_contents = original_value
        else:
            result = compile_and_call(fn, *args, **kwargs)
        # If terminated, this line won't run
        # We can check whether this value is 1 to see if the compiled function finished without terminating
        Num._from_place_(BlockPlace(-1, 0))._set_(Num(1))
        result_type = type(result)
        target = result_type._from_place_(BlockPlace(-2, 0))
        if result_type._is_value_type_():
            target._set_(result)
        else:
            target._copy_from_(result)
        return result

    for read_closure_from_rom, passes in itertools.product((False, True), optimization_levels):
        try:
            cfg, rom_values = compile_fn(run_compiled_with_closure_from_rom if read_closure_from_rom else run_compiled)
        except CompilationError as e:
            assert exception is not None
            while isinstance(e, CompilationError) and e.__cause__ is not None:
                e = e.__cause__
            assert str(e) == str(exception)  # noqa: PT017
            assert type(e) is type(exception)  # noqa: PT017
            raise exception from None

        cfg = run_passes(cfg, passes, OptimizerConfig())
        entry = cfg_to_engine_node(cfg)
        interpreter = Interpreter()
        interpreter.blocks[PlayBlock.EngineRom] = rom_values

        num_result = interpreter.run(entry)
        if exception is None:
            if result_type == Num:
                assert num_result == regular_result
            else:
                assert num_result == 0
        compiled_result = result_type._from_list_(
            [interpreter.get(-2, i) for i in range(result_type._size_())]
        )._as_py_()
        compiled_terminated = interpreter.get(-1, 0) != 1

        if exception is not None:
            assert compiled_terminated, "Compiled function should terminate if regular function raises exception"
            raise exception

        assert compiled_result == regular_result
        assert interpreter.log == log_entries

    return regular_result


def run_compiled[**P](fn: Callable[P, Num], *args: P.args, **kwargs: P.kwargs) -> Num:
    """Runs a function as a compiled function and returns the result."""

    @meta_fn
    def wrapper():
        return compile_and_call(fn, *args, **kwargs)

    results = []
    initial_random_state = random.getstate()
    for passes in optimization_levels:
        random.setstate(initial_random_state)
        cfg, rom_values = compile_fn(wrapper)
        cfg = run_passes(cfg, passes, OptimizerConfig())
        entry = cfg_to_engine_node(cfg)
        interpreter = Interpreter()
        interpreter.blocks[PlayBlock.EngineRom] = rom_values
        result = interpreter.run(entry)
        results.append(result)

    if len(set(results)) != 1:
        raise ValueError(f"Compiled results differ between optimization levels: {results}")

    return results[0]


def implies(a: bool, b: bool) -> bool:
    if a:
        return b
    return True


def is_close(a: float | Vec2, b: float | Vec2, rel_tol: float = 1e-8, abs_tol: float = 1e-8) -> bool:
    if isinstance(a, Vec2):
        return is_close(a.x, b.x, rel_tol, abs_tol) and is_close(a.y, b.y, rel_tol, abs_tol)
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)



from sonolus.backend.place import BlockPlace
from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.value import BackingSource, Value
from sonolus.script.num import Num, _is_num


@meta_fn
def _deref[T: Value](block: int, offset: int, type_: type[T]) -> T:
    block = Num._accept_(block)
    offset = Num._accept_(offset)
    type_ = validate_value(type_)._as_py_()
    if block._is_py_():
        block = block._as_py_()
        if not isinstance(block, int):
            raise TypeError("block must be an integer")
        block = ctx().blocks(block)
    else:
        if not _is_num(block):
            raise TypeError("block must be a Num")
        block = block.index()
    if offset._is_py_():
        offset = offset._as_py_()
        if not isinstance(offset, int):
            raise TypeError("offset must be an integer")
    else:
        if not _is_num(offset):
            raise TypeError("offset must be a Num")
        offset = offset.index()
    if not (isinstance(type_, type) and issubclass(type_, Value)):
        raise TypeError("type_ must be a Value")
    return type_._from_place_(BlockPlace(block, offset))


@meta_fn
def _backing_deref[T: Value](source: BackingSource, type_: type[T]) -> T:
    type_ = validate_value(type_)._as_py_()
    if not isinstance(type_, type) or not issubclass(type_, Value):
        raise TypeError("type_ must be a Value")
    return type_._from_backing_source_(source)



# type: ignore
from __future__ import annotations

import operator
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, Self, TypeGuard, final, runtime_checkable

from sonolus.backend.blocks import BlockData
from sonolus.backend.ir import IRConst, IRExpr, IRGet, IRPureInstr, IRSet
from sonolus.backend.ops import Op
from sonolus.backend.place import BlockPlace
from sonolus.script.internal.context import ctx
from sonolus.script.internal.error import InternalError
from sonolus.script.internal.impl import meta_fn
from sonolus.script.internal.value import BackingValue, DataValue, ExprBackingValue, Value


class _NumMeta(type):
    def __instancecheck__(cls, instance):
        return isinstance(instance, float | int | bool) or _is_num(instance)


def _is_num(value: Any) -> TypeGuard[Num]:
    """Check if a value is a precisely Num instance."""
    return type.__instancecheck__(Num, value)  # type: ignore # noqa: PLC2801


@final
class _Num(Value, metaclass=_NumMeta):
    # This works for ints, floats, and bools
    # Since we don't support complex numbers, real is equal to the original number
    __match_args__ = ("real",)

    data: DataValue

    def __init__(self, data: DataValue | IRExpr):
        if isinstance(data, complex):
            raise TypeError("Cannot create a Num from a complex number")
        if isinstance(data, int):
            data = float(data)
        if isinstance(data, IRConst | IRPureInstr | IRGet):
            data = ExprBackingValue(data)
        if _is_num(data):
            raise InternalError("Cannot create a Num from a Num")
        if not isinstance(data, BlockPlace | BackingValue | float | int | bool):
            raise TypeError(f"Cannot create a Num from {type(data)}")
        self.data = data

    def __str__(self) -> str:
        if isinstance(self.data, float) and self.data.is_integer():
            return str(int(self.data))
        return str(self.data)

    def __repr__(self) -> str:
        return f"Num({self.data})"

    @classmethod
    def _is_concrete_(cls) -> bool:
        return True

    @classmethod
    def _size_(cls) -> int:
        return 1

    @classmethod
    def _is_value_type_(cls) -> bool:
        return True

    @classmethod
    def _from_place_(cls, place: BlockPlace) -> Self:
        return cls(place)

    @classmethod
    def _accepts_(cls, value: Value) -> bool:
        return isinstance(value, Num | float | int | bool)

    @classmethod
    def _accept_(cls, value: Any) -> Self:
        if not cls._accepts_(value):
            raise TypeError(f"Cannot accept {value}")
        if _is_num(value):
            return value
        return cls(value)

    def _is_rom_constant(self) -> bool:
        return (
            ctx()
            and isinstance(self.data, BlockPlace)
            and self.data.block == ctx().blocks.EngineRom
            and isinstance(self.data.index, int)
        )

    def _is_py_(self) -> bool:
        return isinstance(self.data, float | int | bool) or self._is_rom_constant()

    def _as_py_(self) -> Any:
        if not self._is_py_():
            raise ValueError("Not a compile time constant Num")
        if self._is_rom_constant():
            return ctx().rom.get_value(self.data.index + self.data.offset)
        if self.data.is_integer():
            return int(self.data)
        return self.data

    @classmethod
    def _from_list_(cls, values: Iterable[DataValue]) -> Self:
        value = next(iter(values))
        return Num(value)

    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue]:
        return [self.data]

    @classmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        return [prefix]

    def _get_(self) -> Self:
        if ctx():
            if isinstance(self.data, BlockPlace):
                ctx().check_readable(self.data)
            place = ctx().alloc(size=1)
            ctx().add_statements(IRSet(place, self.ir()))
            return Num(place)
        else:
            return Num(self.data)

    def _get_readonly_(self) -> Self:
        if ctx():
            if isinstance(self.data, BlockPlace):
                ctx().check_readable(self.data)
                if isinstance(self.data.block, BlockData) and not ctx().is_writable(self.data):
                    # This block is immutable in the current callback, so no need to copy it in case it changes.
                    return Num(self.data)
            place = ctx().alloc(size=1)
            ctx().add_statements(IRSet(place, self.ir()))
            return Num(place)
        else:
            return Num(self.data)

    def _set_(self, value: Any):
        value = Num._accept_(value)
        if ctx():
            match self.data:
                case BackingValue():
                    self.data.write(value.ir())
                case BlockPlace():
                    ctx().check_writable(self.data)
                    ctx().add_statements(IRSet(self.data, value.ir()))
                case _:
                    raise ValueError("Cannot set a read-only value")
        else:
            self.data = value.data

    def _copy_from_(self, value: Any):
        raise ValueError("Cannot assign to a number")

    def _copy_(self) -> Self:
        if ctx():
            return self._get_()
        else:
            return Num(self.data)

    @classmethod
    def _alloc_(cls) -> Self:
        if ctx():
            return Num(ctx().alloc(size=1))
        else:
            return Num(-1)

    @classmethod
    def _zero_(cls) -> Self:
        if ctx():
            result_place = ctx().alloc(size=1)
            ctx().add_statements(IRSet(result_place, IRConst(0)))
            return cls(result_place)
        else:
            return cls(0)

    def ir(self):
        match self.data:
            case BlockPlace():
                return IRGet(self.data)
            case BackingValue():
                return self.data.read()
            case _:
                return IRConst(self.data)

    def index(self) -> int | BlockPlace:
        if isinstance(self.data, BlockPlace):
            return self._get_().data
        return self.data

    def _bin_op(self, other: Self, const_fn: Callable[[Self, Self], Self | None], ir_op: Op) -> Self:
        if not Num._accepts_(other):
            return NotImplemented
        other = Num._accept_(other)
        const_value = const_fn(self, other)
        if const_value is not None:
            return const_value
        if ctx():
            result_place = ctx().alloc(size=1)
            ctx().add_statements(IRSet(result_place, IRPureInstr(ir_op, [self.ir(), other.ir()])))
            return Num(result_place)
        else:
            raise InternalError("Unexpected call on non-comptime Num instance outside a compilation context")

    def _unary_op(self, py_fn: Callable[[float], float], ir_op: Op) -> Self:
        if self._is_py_():
            return Num(py_fn(self._as_py_()))
        elif ctx():
            result_place = ctx().alloc(size=1)
            ctx().add_statements(IRSet(result_place, IRPureInstr(ir_op, [self.ir()])))
            return Num(result_place)
        else:
            raise InternalError("Unexpected call on non-comptime Num instance outside a compilation context")

    @meta_fn
    def __eq__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() == b._as_py_())
            if a.data == b.data:
                return Num(True)
            return None

        return self._bin_op(other, const_fn, Op.Equal)

    def __hash__(self):
        if self._is_py_():
            return hash(self._as_py_())
        raise ValueError("Cannot hash non compile time constant Num")

    @meta_fn
    def __ne__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() != b._as_py_())
            if a.data == b.data:
                return Num(False)
            return None

        return self._bin_op(other, const_fn, Op.NotEqual)

    @meta_fn
    def __lt__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() < b._as_py_())
            if a.data == b.data:
                return Num(False)
            return None

        return self._bin_op(other, const_fn, Op.Less)

    @meta_fn
    def __le__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() <= b._as_py_())
            if a.data == b.data:
                return Num(True)
            return None

        return self._bin_op(other, const_fn, Op.LessOr)

    @meta_fn
    def __gt__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() > b._as_py_())
            if a.data == b.data:
                return Num(False)
            return None

        return self._bin_op(other, const_fn, Op.Greater)

    @meta_fn
    def __ge__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() >= b._as_py_())
            if a.data == b.data:
                return Num(True)
            return None

        return self._bin_op(other, const_fn, Op.GreaterOr)

    @meta_fn
    def __add__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() + b._as_py_())
            if a._is_py_() and a._as_py_() == 0:
                return b
            if b._is_py_() and b._as_py_() == 0:
                return a
            return None

        return self._bin_op(other, const_fn, Op.Add)

    @meta_fn
    def __sub__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() - b._as_py_())
            if a._is_py_() and a._as_py_() == 0:
                return -b
            if b._is_py_() and b._as_py_() == 0:
                return a
            return None

        return self._bin_op(other, const_fn, Op.Subtract)

    @meta_fn
    def __mul__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() * b._as_py_())
            if a._is_py_():
                if a._as_py_() == 0:
                    return Num(0)
                if a._as_py_() == 1:
                    return b
            if b._is_py_():
                if b._as_py_() == 0:
                    return Num(0)
                if b._as_py_() == 1:
                    return a
            return None

        return self._bin_op(other, const_fn, Op.Multiply)

    @meta_fn
    def __truediv__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                if b._as_py_() == 0:
                    return None
                return Num(a._as_py_() / b._as_py_())
            if b._is_py_():
                if b._as_py_() == 1:
                    return a
                if b._as_py_() == -1:
                    return -a
            return None

        return self._bin_op(other, const_fn, Op.Divide)

    @meta_fn
    def __floordiv__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                if b._as_py_() == 0:
                    return None
                return Num(a._as_py_() // b._as_py_())
            if b._is_py_():
                if b._as_py_() == 1:
                    return a
                if b._as_py_() == -1:
                    return -a
            return None

        return self._bin_op(other, const_fn, Op.Divide)._unary_op(lambda x: x, Op.Floor)

    @meta_fn
    def __mod__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                if b._as_py_() == 0:
                    return None
                return Num(a._as_py_() % b._as_py_())
            return None

        return self._bin_op(other, const_fn, Op.Mod)

    @meta_fn
    def __pow__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                try:
                    return Num(a._as_py_() ** b._as_py_())
                except OverflowError:
                    return None
            if b._is_py_():
                if b._as_py_() == 0:
                    return Num(1)
                if b._as_py_() == 1:
                    return a
            return None

        return self._bin_op(other, const_fn, Op.Power)

    @meta_fn
    def __neg__(self) -> Self:
        return self._unary_op(operator.neg, Op.Negate)

    @meta_fn
    def __pos__(self) -> Self:
        return self

    @meta_fn
    def __abs__(self) -> Self:
        return self._unary_op(abs, Op.Abs)

    @meta_fn
    def __bool__(self):
        if ctx():
            return self != 0
        else:
            if self._is_py_():
                return bool(self._as_py_())
            raise ValueError("Cannot convert non compile time constant Num to bool")

    def and_(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() and b._as_py_())
            if a._is_py_():
                if a._as_py_() == 0:
                    return a
                else:
                    return b
            return None

        return self._bin_op(other, const_fn, Op.And)

    def or_(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() or b._as_py_())
            if a._is_py_():
                if a._as_py_() == 0:
                    return b
                else:
                    return a
            return None

        return self._bin_op(other, const_fn, Op.Or)

    def not_(self) -> Self:
        return self._unary_op(operator.not_, Op.Not)

    @property
    def real(self) -> Self:
        return self

    @property
    def imag(self) -> Self:
        return 0


if TYPE_CHECKING:
    from typing import Protocol

    @runtime_checkable
    class Num(Protocol, int, bool, float):
        def __add__(self, other: Any, /) -> Num | int | bool | float: ...
        def __sub__(self, other: Any, /) -> Num | int | bool | float: ...
        def __mul__(self, other: Any, /) -> Num | int | bool | float: ...
        def __truediv__(self, other: Any, /) -> Num | int | bool | float: ...
        def __floordiv__(self, other: Any, /) -> Num | int | bool | float: ...
        def __mod__(self, other: Any, /) -> Num | int | bool | float: ...
        def __pow__(self, other: Any, /) -> Num | int | bool | float: ...

        def __neg__(self, /) -> Num | int | bool | float: ...
        def __pos__(self, /) -> Num | int | bool | float: ...
        def __abs__(self, /) -> Num | int | bool | float: ...

        def __eq__(self, other: Any, /) -> bool: ...
        def __ne__(self, other: Any, /) -> bool: ...
        def __lt__(self, other: Any, /) -> bool: ...
        def __le__(self, other: Any, /) -> bool: ...
        def __gt__(self, other: Any, /) -> bool: ...
        def __ge__(self, other: Any, /) -> bool: ...

        def __hash__(self, /) -> int: ...

        @property
        def real(self) -> Num | int | bool | float: ...

        @property
        def imag(self) -> Num | int | bool | float: ...
else:
    # Need to do this to satisfy type checkers (especially Pycharm)
    _Num.__name__ = "Num"
    _Num.__qualname__ = "Num"
    globals()["Num"] = _Num



from collections import deque

from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet, IRStmt
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_preorder
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig
from sonolus.backend.place import BlockPlace, SSAPlace, TempBlock

type HasLiveness = SSAPlace | TempBlock


class LivenessAnalysis(CompilerPass):
    def destroys(self) -> set[CompilerPass]:
        return set()

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        self.preprocess(entry)
        self.process(entry)
        return entry

    def preprocess(self, entry: BasicBlock):
        for block in traverse_cfg_preorder(entry):
            block.live_out = None
            block.live_in = set()
            block.live_phi_targets = set()
            block.array_defs_in = set()
            block.array_defs_out = None
            for statement in block.statements:
                statement.live = set()
                statement.visited = False
                statement.uses = self.get_uses(statement, set())
                statement.defs = self.get_defs(statement)
                statement.is_array_init = False  # True if this may be the first assignment to an array
                statement.array_defs = self.get_array_defs(statement)
            block.test.live = set()
            block.test.uses = self.get_uses(block.test, set())
        self.preprocess_arrays(entry)

    def process(self, entry: BasicBlock):
        queue = deque(self.get_exits(entry))
        if not queue:
            raise ValueError("Infinite loop detected")
        while queue:
            block = queue.popleft()
            updated_blocks = self.process_block(block)
            queue.extend(updated_blocks)

    def preprocess_arrays(self, entry: BasicBlock):
        queue = {entry}
        visited = set()
        while queue:
            block = queue.pop()
            array_defs = block.array_defs_in.copy()
            is_first_visit = block not in visited
            visited.add(block)
            for statement in block.statements:
                if statement.array_defs - array_defs:
                    statement.is_array_init = True
                    array_defs.update(statement.array_defs)
                else:
                    statement.is_array_init = False
            if is_first_visit or array_defs != block.array_defs_out:
                block.array_defs_out = array_defs
                for edge in block.outgoing:
                    queue.add(edge.dst)
                    edge.dst.array_defs_in.update(array_defs)

    def process_block(self, block: BasicBlock) -> list[BasicBlock]:
        if block.live_out is None:
            block.live_out = set()
        live: set[HasLiveness] = {
            place
            for place in block.live_out
            if not (isinstance(place, TempBlock) and place.size > 1 and place not in block.array_defs_out)
        }
        block.test.live.update(live)
        live.update(block.test.uses)
        for statement in reversed(block.statements):
            statement.live.update(live)
            if self.can_skip(statement, live):
                continue
            live.difference_update(statement.defs)
            if statement.is_array_init:
                live.difference_update(statement.array_defs)
            live.update(statement.uses)
        prev_sizes_by_block = {
            edge.src: len(edge.src.live_out) if edge.src.live_out is not None else -1 for edge in block.incoming
        }
        live_phi_targets = set()
        for target, args in block.phis.items():
            if target not in live:
                continue
            live.remove(target)
            for src_block, arg in args.items():
                if src_block.live_out is None:
                    src_block.live_out = set()
                src_block.live_out.add(arg)
            live_phi_targets.add(target)
        block.live_in = live
        block.live_phi_targets = live_phi_targets
        updated_blocks = []
        for edge in block.incoming:
            if edge.src.live_out is None:
                edge.src.live_out = set()
            edge.src.live_out.update(live)
            if len(edge.src.live_out) != prev_sizes_by_block[edge.src]:
                updated_blocks.append(edge.src)
        return updated_blocks

    def get_uses(
        self, stmt: IRStmt | BlockPlace | SSAPlace | TempBlock | int, uses: set[HasLiveness]
    ) -> set[HasLiveness]:
        match stmt:
            case IRPureInstr(op=_, args=args) | IRInstr(op=_, args=args):
                for arg in args:
                    self.get_uses(arg, uses)
            case IRGet(place=place):
                self.get_uses(place, uses)
            case IRSet(place=place, value=value):
                if isinstance(place, BlockPlace):
                    if not isinstance(place.block, TempBlock):
                        self.get_uses(place.block, uses)
                    self.get_uses(place.index, uses)
                self.get_uses(value, uses)
            case IRConst() | int():
                pass
            case BlockPlace(block=block, index=index, offset=_):
                self.get_uses(block, uses)
                self.get_uses(index, uses)
            case SSAPlace() | TempBlock():
                uses.add(stmt)
            case _:
                raise TypeError(f"Unexpected statement type: {type(stmt)}")
        return uses

    def get_defs(self, stmt: IRStmt | BlockPlace | SSAPlace | TempBlock | int) -> set[HasLiveness]:
        match stmt:
            case IRSet(place=place, value=_):
                match place:
                    case SSAPlace():
                        return {place}
                    case BlockPlace(block=TempBlock() as temp_block, index=_, offset=_) if temp_block.size == 1:
                        return {temp_block}
        return set()

    def get_array_defs(self, stmt: IRStmt | BlockPlace | SSAPlace | TempBlock | int) -> set[HasLiveness]:
        match stmt:
            case IRSet(place=place, value=_):
                match place:
                    case BlockPlace(block=TempBlock() as temp_block, index=_, offset=_) if temp_block.size > 1:
                        return {temp_block}
        return set()

    def can_skip(self, stmt: IRStmt, live: set[HasLiveness]) -> bool:
        match stmt:
            case IRSet(place=_, value=value):
                if isinstance(value, IRInstr) and value.op.side_effects:
                    return False
                defs = stmt.defs | stmt.array_defs
                return defs and not (defs & live)
        return False

    def get_exits(self, entry: BasicBlock) -> list[BasicBlock]:
        return [block for block in traverse_cfg_preorder(entry) if not block.outgoing]

    def __eq__(self, other):
        return isinstance(other, LivenessAnalysis)

    def __hash__(self):
        return hash(LivenessAnalysis)


def get_live(stmt: IRStmt) -> set[HasLiveness]:
    return stmt.live


def get_live_phi_targets(block: BasicBlock) -> set[HasLiveness]:
    return block.live_phi_targets



# ruff: noqa: E741
import itertools
from math import pi

from hypothesis import assume, given
from hypothesis import strategies as st

from sonolus.script.quad import Quad, Rect
from sonolus.script.vec import Vec2
from tests.script.conftest import is_close, run_and_validate

floats = st.floats(min_value=-9, max_value=9, allow_nan=False, allow_infinity=False)
nonzero_floats = floats.filter(lambda x: abs(x) > 1e-2)


@st.composite
def vecs(draw):
    x = draw(floats)
    y = draw(floats)
    return Vec2(x, y)


def _area(points: list[Vec2]) -> float:
    return 0.5 * abs(sum(p.x * q.y - p.y * q.x for p, q in zip(points, points[1:] + points[:1], strict=True)))


@st.composite
def quads(draw):
    points = [draw(vecs()) for _ in range(4)]
    for p1, p2 in itertools.combinations(points, 2):
        assume((p1 - p2).magnitude > 1e-2)
    centroid = sum(points, Vec2(0, 0)) / 4
    points = sorted(points, key=lambda p: (p - centroid).angle)
    assume(_area(points) > 1e-2)
    return Quad(*points)


@st.composite
def rects(draw):
    l = draw(floats)
    r = draw(floats)
    if l > r:
        l, r = r, l

    b = draw(floats)
    t = draw(floats)
    if b > t:
        b, t = t, b

    return Rect(t=t, r=r, b=b, l=l)


@given(
    quad=quads(),
    translation=vecs(),
)
def test_quad_translate(quad, translation):
    def fn():
        return quad.translate(translation)

    result = run_and_validate(fn)
    assert is_close(result.bl, quad.bl + translation)
    assert is_close(result.tl, quad.tl + translation)
    assert is_close(result.tr, quad.tr + translation)
    assert is_close(result.br, quad.br + translation)


@given(
    quad=quads(),
    factor=vecs(),
)
def test_quad_scale(quad, factor):
    def fn():
        return quad.scale(factor)

    result = run_and_validate(fn)
    assert is_close(result.bl, quad.bl * factor)
    assert is_close(result.tl, quad.tl * factor)
    assert is_close(result.tr, quad.tr * factor)
    assert is_close(result.br, quad.br * factor)


@given(
    quad=quads(),
    factor=vecs(),
    pivot=vecs(),
)
def test_quad_scale_about(quad, factor, pivot):
    def fn():
        return quad.scale_about(factor, pivot)

    result = run_and_validate(fn)
    assert is_close(result.bl, (quad.bl - pivot) * factor + pivot)
    assert is_close(result.tl, (quad.tl - pivot) * factor + pivot)
    assert is_close(result.tr, (quad.tr - pivot) * factor + pivot)
    assert is_close(result.br, (quad.br - pivot) * factor + pivot)


@given(
    quad=quads(),
    angle=floats,
)
def test_quad_rotate(quad, angle):
    def fn():
        return quad.rotate(angle)

    result = run_and_validate(fn)
    assert is_close(result.bl, quad.bl.rotate(angle))
    assert is_close(result.tl, quad.tl.rotate(angle))
    assert is_close(result.tr, quad.tr.rotate(angle))
    assert is_close(result.br, quad.br.rotate(angle))


@given(
    quad=quads(),
    angle=floats,
    pivot=vecs(),
)
def test_quad_rotate_about(quad, angle, pivot):
    def fn():
        return quad.rotate_about(angle, pivot)

    result = run_and_validate(fn)
    assert is_close(result.bl, quad.bl.rotate_about(angle, pivot))
    assert is_close(result.tl, quad.tl.rotate_about(angle, pivot))
    assert is_close(result.tr, quad.tr.rotate_about(angle, pivot))
    assert is_close(result.br, quad.br.rotate_about(angle, pivot))


@given(quad=quads())
def test_quad_center(quad):
    def fn():
        return quad.center

    result = run_and_validate(fn)
    expected = (quad.bl + quad.tr + quad.tl + quad.br) / 4
    assert is_close(result, expected)


def test_quad_permute():
    for rotation in range(-5, 6):
        quad = Quad(
            Vec2(-1, -1),
            Vec2(-1, 1),
            Vec2(1, 1),
            Vec2(1, -1),
        )

        def fn():
            return quad.permute(rotation)  # noqa: B023

        result = run_and_validate(fn)
        expected = quad.rotate(rotation * pi / 2)
        assert is_close(result.bl, expected.bl)
        assert is_close(result.tl, expected.tl)
        assert is_close(result.tr, expected.tr)
        assert is_close(result.br, expected.br)


# Rect Tests
@given(
    rect=rects(),
    translation=vecs(),
)
def test_rect_translate(rect, translation):
    def fn():
        return rect.translate(translation)

    result = run_and_validate(fn)
    assert is_close(result.t, rect.t + translation.y)
    assert is_close(result.r, rect.r + translation.x)
    assert is_close(result.b, rect.b + translation.y)
    assert is_close(result.l, rect.l + translation.x)


@given(
    rect=rects(),
    factor=vecs(),
)
def test_rect_scale(rect, factor):
    def fn():
        return rect.scale(factor)

    result = run_and_validate(fn)
    assert is_close(result.t, rect.t * factor.y)
    assert is_close(result.r, rect.r * factor.x)
    assert is_close(result.b, rect.b * factor.y)
    assert is_close(result.l, rect.l * factor.x)


@given(
    rect=rects(),
    factor=vecs(),
    pivot=vecs(),
)
def test_rect_scale_about(rect, factor, pivot):
    def fn():
        return rect.scale_about(factor, pivot)

    result = run_and_validate(fn)
    assert is_close(result.t, (rect.t - pivot.y) * factor.y + pivot.y)
    assert is_close(result.r, (rect.r - pivot.x) * factor.x + pivot.x)
    assert is_close(result.b, (rect.b - pivot.y) * factor.y + pivot.y)
    assert is_close(result.l, (rect.l - pivot.x) * factor.x + pivot.x)


@given(
    rect=rects(),
    expansion=vecs(),
)
def test_rect_expand(rect, expansion):
    def fn():
        return rect.expand(expansion)

    result = run_and_validate(fn)
    assert is_close(result.t, rect.t + expansion.y)
    assert is_close(result.r, rect.r + expansion.x)
    assert is_close(result.b, rect.b - expansion.y)
    assert is_close(result.l, rect.l - expansion.x)


@given(
    rect=rects(),
    point=vecs(),
)
def test_rect_contains_point(rect, point):
    def fn():
        return rect.contains_point(point)

    result = run_and_validate(fn)
    expected = rect.l <= point.x <= rect.r and rect.b <= point.y <= rect.t
    assert is_close(result, expected)


@given(
    center=vecs(),
    dimensions=vecs(),
)
def test_rect_from_center(center, dimensions):
    def fn():
        return Rect.from_center(center, dimensions)

    result = run_and_validate(fn)
    assert is_close(result.center.x, center.x)
    assert is_close(result.center.y, center.y)
    assert is_close(result.w, dimensions.x)
    assert is_close(result.h, dimensions.y)


@given(rect=rects())
def test_rect_as_quad(rect):
    def fn():
        return rect.as_quad()

    result = run_and_validate(fn)
    assert is_close(result.bl, rect.bl)
    assert is_close(result.tl, rect.tl)
    assert is_close(result.tr, rect.tr)
    assert is_close(result.br, rect.br)


@st.composite
def quad_and_point(draw):
    quad = draw(quads())
    min_x = min(quad.bl.x, quad.tl.x, quad.tr.x, quad.br.x)
    max_x = max(quad.bl.x, quad.tl.x, quad.tr.x, quad.br.x)
    min_y = min(quad.bl.y, quad.tl.y, quad.tr.y, quad.br.y)
    max_y = max(quad.bl.y, quad.tl.y, quad.tr.y, quad.br.y)
    x = draw(st.floats(min_value=min_x, max_value=max_x))
    y = draw(st.floats(min_value=min_y, max_value=max_y))
    point = Vec2(x, y)
    return quad, point


@given(quad_point=quad_and_point())
def test_quad_contains_point(quad_point):
    quad, point = quad_point

    def fn():
        return quad.contains_point(point)

    # Don't have an easy way to validate the correct solution
    # But this still checks that the compiled and Python versions are consistent
    run_and_validate(fn)



from pydori.lib.skin import Skin
from sonolus.script.bucket import Bucket, JudgmentWindow, bucket, bucket_sprite, buckets
from sonolus.script.interval import Interval
from sonolus.script.runtime import level_score
from sonolus.script.text import StandardText


@buckets
class Buckets:
    tap_note: Bucket = bucket(
        sprites=[
            bucket_sprite(
                sprite=Skin.tap_note,
                x=0,
                y=0,
                w=2,
                h=2,
                rotation=-90,
            ),
        ],
        unit=StandardText.MILLISECOND_UNIT,
    )
    hold_head_note: Bucket = bucket(
        sprites=[
            bucket_sprite(
                sprite=Skin.hold_connector,
                x=0.5,
                y=0,
                w=2,
                h=5,
                rotation=-90,
            ),
            bucket_sprite(
                sprite=Skin.hold_head_note,
                x=-2,
                y=0,
                w=2,
                h=2,
                rotation=-90,
            ),
        ],
        unit=StandardText.MILLISECOND_UNIT,
    )
    hold_end_note: Bucket = bucket(
        sprites=[
            bucket_sprite(
                sprite=Skin.hold_connector,
                x=-0.5,
                y=0,
                w=2,
                h=5,
                rotation=-90,
            ),
            bucket_sprite(
                sprite=Skin.hold_end_note,
                x=2,
                y=0,
                w=2,
                h=2,
                rotation=-90,
            ),
        ],
        unit=StandardText.MILLISECOND_UNIT,
    )
    hold_tick_note: Bucket = bucket(
        sprites=[
            bucket_sprite(
                sprite=Skin.hold_connector,
                x=0,
                y=0,
                w=2,
                h=5.5,
                rotation=-90,
            ),
            bucket_sprite(
                sprite=Skin.hold_tick_note,
                x=0,
                y=0,
                w=2,
                h=2,
                rotation=-90,
            ),
        ],
        unit=StandardText.MILLISECOND_UNIT,
    )
    flick_note: Bucket = bucket(
        sprites=[
            bucket_sprite(
                sprite=Skin.flick_note,
                x=0,
                y=0,
                w=2,
                h=2,
                rotation=-90,
            ),
            bucket_sprite(
                sprite=Skin.flick_arrow,
                x=1,
                y=0,
                w=2,
                h=2,
                rotation=-90,
            ),
        ],
        unit=StandardText.MILLISECOND_UNIT,
    )
    directional_flick_note: Bucket = bucket(
        sprites=[
            bucket_sprite(
                sprite=Skin.right_flick_note,
                x=2,
                y=0,
                w=2,
                h=2,
                rotation=-90,
            ),
            bucket_sprite(
                sprite=Skin.left_flick_note,
                x=-2,
                y=0,
                w=2,
                h=2,
                rotation=90,
            ),
            bucket_sprite(
                sprite=Skin.right_flick_arrow,
                x=3,
                y=0,
                w=2,
                h=2,
                rotation=-90,
            ),
            bucket_sprite(
                sprite=Skin.left_flick_arrow,
                x=-3,
                y=0,
                w=2,
                h=2,
                rotation=90,
            ),
        ],
        unit=StandardText.MILLISECOND_UNIT,
    )


note_judgment_window = JudgmentWindow(
    perfect=Interval(-0.05, 0.05),
    great=Interval(-0.1, 0.1),
    good=Interval(-0.15, 0.15),
)


def init_buckets():
    # Multiply by 1000 so buckets are in milliseconds.
    Buckets.tap_note.window @= note_judgment_window * 1000
    Buckets.hold_head_note.window @= note_judgment_window * 1000
    Buckets.hold_end_note.window @= note_judgment_window * 1000
    Buckets.hold_tick_note.window @= note_judgment_window * 1000
    Buckets.flick_note.window @= note_judgment_window * 1000
    Buckets.directional_flick_note.window @= note_judgment_window * 1000


def init_score():
    level_score().update(
        perfect_multiplier=1,
        great_multiplier=0.5,
        good_multiplier=0.25,
    )



from sonolus.backend.ops import Op
from sonolus.backend.place import BlockPlace, Place, SSAPlace

type IRExpr = IRConst | IRPureInstr | IRGet
type IRStmt = IRExpr | IRInstr | IRSet


class IRConst:
    value: float | int

    def __init__(self, value: float):
        if isinstance(value, bool):
            value = int(value)
        self.value = value

    def __repr__(self):
        return f"IRConst({self.value!r})"

    def __str__(self):
        return f"{self.value}"

    def __eq__(self, other):
        return isinstance(other, IRConst) and self.value == other.value

    def __hash__(self):
        return hash(self.value)


class IRPureInstr:
    op: Op
    args: list[IRExpr]

    def __init__(self, op: Op, args: list[IRExpr]):
        assert op.pure
        self.op = op
        self.args = args

    def __repr__(self):
        return f"IRPureInstr({self.op!r}, {self.args!r})"

    def __str__(self):
        return f"{self.op.name}({', '.join(map(str, self.args))})"

    def __eq__(self, other):
        return isinstance(other, IRPureInstr) and self.op == other.op and self.args == other.args

    def __hash__(self):
        return hash((self.op, tuple(self.args)))


class IRInstr:
    op: Op
    args: list[IRExpr]

    def __init__(self, op: Op, args: list[IRExpr]):
        self.op = op
        self.args = args

    def __repr__(self):
        return f"IRInstr({self.op!r}, {self.args!r})"

    def __str__(self):
        return f"{self.op.name}({', '.join(map(str, self.args))})"

    def __eq__(self, other):
        return isinstance(other, IRInstr) and self.op == other.op and self.args == other.args

    def __hash__(self):
        return hash((self.op, tuple(self.args)))


class IRGet:
    place: Place

    def __init__(self, place: Place):
        self.place = place

    def __repr__(self):
        return f"IRGet({self.place!r})"

    def __str__(self):
        return f"{self.place}"

    def __eq__(self, other):
        return isinstance(other, IRGet) and self.place == other.place

    def __hash__(self):
        return hash(self.place)


class IRSet:
    place: Place
    value: IRExpr | IRInstr

    def __init__(self, place: Place, value: IRExpr | IRInstr):
        self.place = place
        self.value = value

    def __repr__(self):
        return f"IRSet({self.place!r}, {self.value!r})"

    def __str__(self):
        match self.place:
            case BlockPlace():
                return f"{self.place} <- {self.value}"
            case SSAPlace():
                return f"{self.place} := {self.value}"
            case _:
                raise TypeError(f"Invalid place: {self.place}")

    def __eq__(self, other):
        return isinstance(other, IRSet) and self.place == other.place and self.value == other.value

    def __hash__(self):
        return hash((self.place, self.value))



import textwrap
from dataclasses import dataclass, field

from sonolus.backend.ops import Op

type EngineNode = ConstantNode | FunctionNode


@dataclass(slots=True)
class ConstantNode:
    value: float
    _hash: int = field(init=False, repr=False)

    def __post_init__(self):
        self._hash = hash(self.value)

    def __hash__(self):
        return hash(self.value)


@dataclass(slots=True)
class FunctionNode:
    func: Op
    args: list[EngineNode]
    _hash: int = field(init=False, repr=False)

    def __post_init__(self):
        self._hash = hash((self.func, tuple(self.args)))

    def __hash__(self):
        return self._hash


def format_engine_node(node: EngineNode) -> str:
    if isinstance(node, ConstantNode):
        return str(node.value)
    elif isinstance(node, FunctionNode):
        match len(node.args):
            case 0:
                return f"{node.func.name}()"
            case 1:
                return f"{node.func.name}({format_engine_node(node.args[0])})"
            case _:
                return f"{node.func.name}(\n{
                    textwrap.indent('\n'.join(format_engine_node(arg) for arg in node.args), '  ')
                }\n)"
    else:
        raise ValueError(f"Invalid engine node: {node}")



from pydori.lib.layer import LAYER_ARROW, LAYER_CONNECTOR, LAYER_NOTE, get_z
from pydori.lib.layout import (
    layout_directional_flick_arrow,
    layout_flick_arrow,
    layout_hold_connector,
    layout_note_body,
    transform_vec,
)
from pydori.lib.note import NoteKind, get_note_arrow_sprite, get_note_body_sprite
from pydori.lib.skin import Skin
from sonolus.script.sprite import Sprite
from sonolus.script.transform import Transform2d
from sonolus.script.vec import Vec2

# Scale factor for tutorial intro notes
INTRO_SCALE = 3

# The y-coordinate tutorial intro notes are drawn at prior to being centered and scaled up
INTRO_DRAW_Y = 1

# Length of connector to show before and/or after hold notes before centering and scaling up
INTRO_CONNECTOR_LENGTH = 1


def draw_tutorial_intro_note(
    kind: NoteKind,
    direction: int = 0,
    lane: float = 0,
    is_hold_flick_end: bool = False,
):
    """Draw the intro for a tutorial note.

    This works by laying out a note normally (like in play mode), then translating and scaling it so it's
    larger and centered on the screen.
    """
    orig_center = transform_vec(Vec2(0, INTRO_DRAW_Y))
    post_transform = (
        Transform2d.new()
        .translate(-orig_center)  # Center the note on the screen
        .scale(Vec2(INTRO_SCALE, INTRO_SCALE))  # Scale it up
    )
    body_sprite = get_note_body_sprite(kind, direction)
    arrow_sprite = get_note_arrow_sprite(kind, direction)
    draw_tutorial_intro_note_body(body_sprite, post_transform, lane)
    match kind:
        case NoteKind.FLICK:
            draw_tutorial_intro_flick_arrow(arrow_sprite, post_transform, lane)
            if is_hold_flick_end:
                draw_tutorial_intro_connector(INTRO_DRAW_Y - INTRO_CONNECTOR_LENGTH, INTRO_DRAW_Y, post_transform, lane)
        case NoteKind.DIRECTIONAL_FLICK:
            draw_tutorial_intro_directional_flick_arrow(arrow_sprite, direction, post_transform, lane)
        case NoteKind.HOLD_HEAD:
            draw_tutorial_intro_connector(INTRO_DRAW_Y, INTRO_DRAW_Y + INTRO_CONNECTOR_LENGTH, post_transform, lane)
        case NoteKind.HOLD_END:
            draw_tutorial_intro_connector(INTRO_DRAW_Y - INTRO_CONNECTOR_LENGTH, INTRO_DRAW_Y, post_transform, lane)
        case NoteKind.HOLD_TICK:
            draw_tutorial_intro_connector(
                INTRO_DRAW_Y - INTRO_CONNECTOR_LENGTH, INTRO_DRAW_Y + INTRO_CONNECTOR_LENGTH, post_transform, lane
            )
        case _:
            pass


def draw_tutorial_intro_note_body(
    sprite: Sprite,
    post_transform: Transform2d,
    lane: float,
):
    original_layout = layout_note_body(lane, INTRO_DRAW_Y)
    layout = post_transform.transform_quad(original_layout)
    sprite.draw(layout, z=get_z(LAYER_NOTE))


def draw_tutorial_intro_flick_arrow(
    sprite: Sprite,
    post_transform: Transform2d,
    lane: float,
):
    original_layout = layout_flick_arrow(lane, INTRO_DRAW_Y, progress=0.5)
    layout = post_transform.transform_quad(original_layout)
    sprite.draw(layout, z=get_z(LAYER_ARROW))


def draw_tutorial_intro_directional_flick_arrow(
    sprite: Sprite,
    direction: int,
    post_transform: Transform2d,
    lane: float,
):
    for i in range(abs(direction)):
        original_layout = layout_directional_flick_arrow(lane, INTRO_DRAW_Y, direction, i, progress=0.5)
        layout = post_transform.transform_quad(original_layout)
        sprite.draw(layout, z=get_z(LAYER_ARROW))


def draw_tutorial_intro_connector(
    y_a: float,
    y_b: float,
    post_transform: Transform2d,
    lane: float,
):
    original_layout = layout_hold_connector(lane, lane, y_a, y_b)
    layout = post_transform.transform_quad(original_layout)
    sprite = Skin.hold_connector
    sprite.draw(layout, z=get_z(LAYER_CONNECTOR))



from pydori.lib.buckets import Buckets
from pydori.lib.effect import Effects
from pydori.lib.particle import Particles
from pydori.lib.skin import Skin
from pydori.play.connector import HoldConnector, SimLine
from pydori.play.event import BpmChange, TimescaleChange
from pydori.play.note import ALL_NOTE_TYPES, HoldManager
from pydori.play.stage import Stage
from sonolus.script.engine import PlayMode

play_mode = PlayMode(
    archetypes=[
        Stage,
        *ALL_NOTE_TYPES,
        HoldManager,
        HoldConnector,
        SimLine,
        BpmChange,
        TimescaleChange,
    ],
    skin=Skin,
    effects=Effects,
    particles=Particles,
    buckets=Buckets,
)



from collections.abc import Callable, Sequence
from contextvars import ContextVar
from typing import Any, Literal, Never

from sonolus.backend.mode import Mode
from sonolus.backend.ops import Op
from sonolus.backend.optimize.flow import cfg_to_mermaid
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig, run_passes
from sonolus.backend.optimize.simplify import RenumberVars
from sonolus.script.internal.context import GlobalContextState, ReadOnlyMemory, ctx, set_ctx
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.native import native_function
from sonolus.script.internal.simulation_context import SimulationContext
from sonolus.script.num import Num

debug_log_callback = ContextVar[Callable[[Num], None]]("debug_log_callback")


@meta_fn
def error(message: str | None = None) -> Never:  # type: ignore
    """Raise an error.

    This function is used to raise an error during runtime.
    When this happens, the game will pause in debug mode. The current callback will also immediately return 0.
    """
    message = validate_value(message)._as_py_() or "Error"  # type: ignore
    if not isinstance(message, str):
        raise ValueError("Expected a string")
    if ctx():
        debug_log(ctx().map_constant(message))
        debug_pause()
        terminate()
    else:
        raise RuntimeError(message)


@meta_fn
def static_error(message: str | None = None) -> Never:
    """Raise a static error.

    This function is used to raise an error during compile-time if the compiler cannot guarantee that
    this function will not be called during runtime.
    """
    message = validate_value(message)._as_py_() or "Error"  # type: ignore
    if not isinstance(message, str):
        raise ValueError("Expected a string")
    raise RuntimeError(message)


@meta_fn
def debug_log(value: int | float | bool):
    """Log a value in debug mode."""
    if debug_log_callback.get(None):
        return debug_log_callback.get()(value)  # type: ignore
    else:
        return _debug_log(value)


@native_function(Op.DebugLog)
def _debug_log(value: int | float | bool):
    print(f"[DEBUG] {value}")
    return 0


@native_function(Op.DebugPause)
def debug_pause():
    """Pause the game if in debug mode."""
    input("[DEBUG] Paused")


def assert_true(value: int | float | bool, message: str | None = None):
    message = message if message is not None else "Assertion failed"
    if not value:
        error(message)


def assert_false(value: int | float | bool, message: str | None = None):
    message = message if message is not None else "Assertion failed"
    if value:
        error(message)


@meta_fn
def assert_unreachable(message: str | None = None) -> Never:
    # This works a bit differently from assert_never from typing in that it throws an error if the Sonolus.py
    # compiler cannot guarantee that this function will not be called, which is different from what type checkers
    # may be able to infer.
    message = validate_value(message)._as_py_() or "Unreachable code reached"  # type: ignore
    raise RuntimeError(message)


@meta_fn
def terminate():
    if ctx():
        set_ctx(ctx().into_dead())
    else:
        raise RuntimeError("Terminated")


def visualize_cfg(
    fn: Callable[[], Any] | Callable[[type], Any],
    /,
    *,
    mode: Mode = Mode.PLAY,
    callback: str = "",
    archetype: type | None = None,
    archetypes: list[type] | None = None,
    passes: Sequence[CompilerPass] | Literal["minimal", "fast", "standard"] = "fast",
) -> str:
    from sonolus.backend.optimize.optimize import FAST_PASSES, MINIMAL_PASSES, STANDARD_PASSES
    from sonolus.build.compile import callback_to_cfg

    match passes:
        case "minimal":
            passes = [
                *MINIMAL_PASSES[:-1],
                RenumberVars(),
            ]
        case "fast":
            passes = [
                *FAST_PASSES[:-1],
                RenumberVars(),
            ]
        case "standard":
            passes = [
                *STANDARD_PASSES[:-1],
                RenumberVars(),
            ]

    global_state = GlobalContextState(
        mode,
        {a: i for i, a in enumerate(archetypes)} if archetypes is not None else None,
        ReadOnlyMemory(),
    )

    cfg = callback_to_cfg(global_state, fn, callback, archetype=archetype)  # type: ignore
    cfg = run_passes(cfg, passes, OptimizerConfig(mode=mode))
    return cfg_to_mermaid(cfg)


def simulation_context() -> SimulationContext:
    return SimulationContext()



from sonolus.backend.ops import Op
from sonolus.script.internal.native import native_function


@native_function(Op.BeatToBPM)
def beat_to_bpm(beat: float) -> float:
    """Get the bpm at the given beat.

    Args:
        beat: The beat to get the bpm at.

    Returns:
        The bpm at the given beat.
    """
    raise NotImplementedError


@native_function(Op.BeatToTime)
def beat_to_time(beat: float) -> float:
    """Get the time at the given beat.

    Args:
        beat: The beat to get the time at.

    Returns:
        The time at the given beat.
    """
    raise NotImplementedError


@native_function(Op.BeatToStartingBeat)
def beat_to_starting_beat(beat: float) -> float:
    """Get the starting beat of the bpm section at the given beat.

    I.e. the beat of the bpm change at or immediately before the given beat.

    Args:
        beat: The beat to get the starting beat of the bpm section at.

    Returns:
        The starting beat of the bpm section at the given beat.
    """
    raise NotImplementedError


@native_function(Op.BeatToStartingTime)
def beat_to_starting_time(beat: float) -> float:
    """Get the starting time of the bpm section at the given beat.

    I.e. the time of the bpm change at or immediately before the given beat.

    Args:
        beat: The beat to get the starting time of the bpm section at.

    Returns:
        The starting time of the bpm section at the given beat.
    """
    raise NotImplementedError


@native_function(Op.TimeToScaledTime)
def time_to_scaled_time(time: float) -> float:
    """Get the scaled (timescale adjusted) time at the given time.

    Args:
        time: The time to get the scaled time at.

    Returns:
        The scaled (timescale adjusted) time at the given time.
    """
    raise NotImplementedError


@native_function(Op.TimeToStartingScaledTime)
def time_to_starting_scaled_time(time: float) -> float:
    """Get the starting scaled (timescale adjusted) time at the given time.

    I.e. the scaled time of the timescale change at or immediately before the given time.

    Args:
        time: The time to get the starting scaled time at.

    Returns:
        The starting scaled time at the given time.
    """
    raise NotImplementedError


@native_function(Op.TimeToStartingTime)
def time_to_starting_time(time: float) -> float:
    """Get the starting time of the timescale section at the given time.

    I.e. the time of the timescale change at or immediately before the given time.

    Args:
        time: The time to get the starting time of the timescale section at.

    Returns:
        The starting time of the timescale section at the given time.
    """
    raise NotImplementedError


@native_function(Op.TimeToTimeScale)
def time_to_timescale(time: float) -> float:
    """Get the timescale at the given time.

    Args:
        time: The time to get the timescale at.

    Returns:
        The timescale at the given time.
    """
    raise NotImplementedError



from __future__ import annotations

import inspect
from abc import ABCMeta
from collections.abc import Iterable
from inspect import getmro
from typing import Any, ClassVar, Self, TypeVar, dataclass_transform, get_origin

from sonolus.backend.place import BlockPlace
from sonolus.script.internal.context import ctx
from sonolus.script.internal.descriptor import SonolusDescriptor
from sonolus.script.internal.generic import (
    GenericValue,
    accept_and_infer_types,
    validate_and_resolve_type,
    validate_concrete_type,
    validate_type_spec,
)
from sonolus.script.internal.impl import meta_fn
from sonolus.script.internal.value import BackingSource, DataValue, Value
from sonolus.script.num import Num


class RecordMeta(ABCMeta):
    @meta_fn
    def __pos__[T](cls: type[T]) -> T:
        """Create a zero-initialized record instance."""
        return cls._zero_()


@dataclass_transform(eq_default=True)
class Record(GenericValue, metaclass=RecordMeta):
    """Base class for user-defined data structures.

    Usage:
        A regular record:
        ```python
        class MyRecord(Record):
            field1: int
            field2: bool
        ```

        A generic record:
        ```python
        class MyGenericRecord[T, U](Record):
            field1: T
            field2: U
        ```

        Creating an instance:
        ```python
        record = MyRecord(field1=42, field2=True)
        record_2 = MyGenericRecord[int, int](field1=42, field2=100)
        record_3 = MyGenericRecord(field1=42, field2=100)  # Type arguments can be inferred
        record_4 = +MyRecord  # Create a zero-initialized record
        record_5 = +MyGenericRecord[int, int]
        ```

        Copying a record:
        ```python
        record_copy = +record
        ```
    """

    _value_: dict[str, Value]
    _fields_: ClassVar[list[_RecordField] | None] = None
    _constructor_signature_: ClassVar[inspect.Signature]

    @classmethod
    def _validate_type_args_(cls, args: tuple[Any, ...]) -> tuple[Any, ...]:
        if cls._fields_ is None:
            raise TypeError("Base Record class cannot have type arguments")
        return super()._validate_type_args_(args)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        is_parameterizing = cls._type_args_ is not None and all(
            getattr(parent, "_type_args_", None) is None for parent in getmro(cls)[1:]
        )
        if is_parameterizing:
            fields = []
            offset = 0
            for generic_field in cls._fields_:
                resolved_type = validate_and_resolve_type(generic_field.type, cls._type_vars_to_args_)
                resolved_type = validate_concrete_type(resolved_type)
                field = _RecordField(generic_field.name, resolved_type, generic_field.index, offset)
                fields.append(field)
                setattr(cls, field.name, field)
                offset += resolved_type._size_()
            cls._fields_ = fields
            return
        is_inheriting_from_existing_record_class = cls._fields_ is not None
        if is_inheriting_from_existing_record_class and not is_parameterizing:
            # The main reason this is disallowed is that subclasses wouldn't be substitutable for their parent classes
            # Assignment of a subclass instance to a variable of the parent class would either be disallowed or would
            # require object slicing. Either way, it could lead to confusion.
            # Dealing with generic supertypes is also tricky, so it isn't really worth the effort to support this.
            raise TypeError("Subclassing of a Record is not supported")

        hints = inspect.get_annotations(cls, eval_str=True)
        fields = []
        params = []
        index = 0
        offset = 0
        for name, hint in hints.items():
            if name not in cls.__annotations__:
                continue
            if hint is ClassVar or get_origin(hint) is ClassVar:
                continue
            if hasattr(cls, name):
                raise TypeError("Default values are not supported for Record fields")
            type_ = validate_type_spec(hint)
            fields.append(_RecordField(name, type_, index, offset))
            if isinstance(type_, type) and issubclass(type_, Value) and type_._is_concrete_():
                offset += type_._size_()
            setattr(cls, name, fields[-1])
            index += 1
            params.append(
                inspect.Parameter(
                    name,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=type_,
                )
            )

        cls._parameterized_ = {}
        cls._fields_ = fields
        cls._constructor_signature_ = inspect.Signature(params)

        _add_inplace_ops(cls)

        cls.__match_args__ = tuple(field.name for field in fields)

        if len(getattr(cls, "__type_params__", ())) == 0:
            # Make the class behave as the parameterized version
            cls._type_args_ = ()
            cls._type_vars_to_args_ = {}
            cls._parameterized_[()] = cls

    def __new__(cls, *args, **kwargs):
        # We override __new__ to allow changing to the parameterized version
        if cls._constructor_signature_ is None:
            raise TypeError(f"Cannot instantiate {cls.__name__}")
        bound = cls._constructor_signature_.bind(*args, **kwargs)
        bound.apply_defaults()
        values = {}
        type_vars = {}
        for field in cls._fields_:
            value = bound.arguments[field.name]
            value = accept_and_infer_types(field.type, value, type_vars)
            values[field.name] = value._get_()
        for type_param in cls.__type_params__:
            if type_param not in type_vars:
                raise TypeError(f"Type parameter {type_param} cannot be inferred and must be provided explicitly")
        type_args = tuple(type_vars[type_param] for type_param in cls.__type_params__)
        if cls._type_args_ is not None:
            parameterized = cls
        else:
            parameterized = cls[type_args]
        result: cls = object.__new__(parameterized)  # type: ignore
        result._value_ = values
        return result

    def __init__(self, *args, **kwargs):
        # Initialization is done in __new__ and other methods
        pass

    @classmethod
    def _raw(cls, **kwargs) -> Self:
        result = object.__new__(cls)
        result._value_ = kwargs
        return result

    @classmethod
    def _size_(cls) -> int:
        return sum(field.type._size_() for field in cls._fields_)

    @classmethod
    def _is_value_type_(cls) -> bool:
        return False

    @classmethod
    def _from_backing_source_(cls, source: BackingSource) -> Self:
        result = object.__new__(cls)
        result._value_ = {
            field.name: field.type._from_backing_source_(
                lambda offset, field_offset=field.offset: source((Num(offset) + Num(field_offset)).ir())  # type: ignore
            )
            for field in cls._fields_
        }
        return result

    @classmethod
    def _from_place_(cls, place: BlockPlace) -> Self:
        result = object.__new__(cls)
        result._value_ = {field.name: field.type._from_place_(place.add_offset(field.offset)) for field in cls._fields_}
        return result

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        return issubclass(type(value), cls)

    @classmethod
    def _accept_(cls, value: Any) -> Self:
        if not cls._accepts_(value):
            raise TypeError(f"Cannot accept value {value} as {cls.__name__}")
        return value

    def _is_py_(self) -> bool:
        return all(value._is_py_() for value in self._value_.values())

    def _as_py_(self) -> Self:
        if not self._is_py_():
            raise ValueError("Not a python value")
        return self

    @classmethod
    def _from_list_(cls, values: Iterable[DataValue]) -> Self:
        iterator = iter(values)
        return cls(**{field.name: field.type._from_list_(iterator) for field in cls._fields_})

    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue | str]:
        result = []
        for field in self._fields_:
            result.extend(self._value_[field.name]._to_list_(level_refs))
        return result

    @classmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        result = []
        for field in cls._fields_:
            result.extend(field.type._flat_keys_(f"{prefix}.{field.name}"))
        if len(result) == 1:
            return [prefix]
        return result

    def _get_(self) -> Self:
        return self

    def _set_(self, value: Any):
        raise TypeError("Record does not support _set_")

    def _copy_from_(self, value: Any):
        value = self._accept_(value)
        for field in self._fields_:
            field.__set__(self, field.__get__(value))

    def _copy_(self) -> Self:
        return type(self)._raw(**{field.name: self._value_[field.name]._copy_() for field in self._fields_})

    @classmethod
    def _alloc_(cls) -> Self:
        # Compared to using the constructor, this avoids unnecessary _get_ calls
        result = object.__new__(cls)
        result._value_ = {field.name: field.type._alloc_() for field in cls._fields_}
        return result

    @classmethod
    def _zero_(cls) -> Self:
        result = object.__new__(cls)
        result._value_ = {field.name: field.type._zero_() for field in cls._fields_}
        return result

    def __str__(self):
        return f"{self.__class__.__name__}({
            ', '.join(f'{field.name}={field.get_internal(self)}' for field in self._fields_)
        })"

    def __repr__(self):
        return f"{self.__class__.__name__}({
            ', '.join(f'{field.name}={field.get_internal(self)!r}' for field in self._fields_)
        })"

    @meta_fn
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False
        result: Num = Num._accept_(True)
        for field in self._fields_:
            result = result.and_(field.__get__(self) == field.__get__(other))
        return result

    @meta_fn
    def __ne__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return True
        result: Num = Num._accept_(False)
        for field in self._fields_:
            result = result.or_(field.__get__(self) != field.__get__(other))
        return result

    def __hash__(self):
        return hash(tuple(field.__get__(self) for field in self._fields_))

    @meta_fn
    def __pos__(self) -> Self:
        """Return a copy of the record."""
        return self._copy_()

    @classmethod
    @meta_fn
    def type_var_value(cls, var: TypeVar, /) -> Any:
        """Return the value of a type variable.

        Args:
            var: The type variable to get the value of.

        Returns:
            The value of the type variable.
        """
        return super().type_var_value(var)


class _RecordField(SonolusDescriptor):
    def __init__(self, name: str, type_: type[Value] | Any, index: int, offset: int):
        self.name = name
        self.type = type_
        self.index = index
        self.offset = offset

    def get_internal(self, instance: Record) -> Value:
        return instance._value_[self.name]

    def __get__(self, instance: Record | None, owner=None):
        if instance is None:
            return self
        result = instance._value_[self.name]._get_readonly_()
        if ctx():
            return result
        else:
            return result._as_py_()

    def __set__(self, instance: Record, value):
        value = self.type._accept_(value)
        if self.type._is_value_type_():
            instance._value_[self.name]._set_(value)
        else:
            instance._value_[self.name]._copy_from_(value)


_ops_to_inplace_ops = {
    "__add__": "__iadd__",
    "__sub__": "__isub__",
    "__mul__": "__imul__",
    "__truediv__": "__itruediv__",
    "__floordiv__": "__ifloordiv__",
    "__mod__": "__imod__",
    "__pow__": "__ipow__",
    "__lshift__": "__ilshift__",
    "__rshift__": "__irshift__",
    "__or__": "__ior__",
    "__xor__": "__ixor__",
    "__and__": "__iand__",
    "__matmul__": "__imatmul__",
}


def _add_inplace_ops(cls):
    for op, inplace_op in _ops_to_inplace_ops.items():
        if hasattr(cls, op) and not hasattr(cls, inplace_op):
            setattr(cls, inplace_op, _make_inplace_op(op))
    return cls


def _make_inplace_op(op: str):
    @meta_fn
    def inplace_op(self, other):
        _compiler_internal_ = True  # noqa: F841
        self._copy_from_(getattr(self, op)(other))
        return self

    return inplace_op



from pydori.lib.connector import draw_hold_connector, draw_sim_line
from pydori.watch.note import WatchNote
from sonolus.script.archetype import EntityRef, WatchArchetype, imported
from sonolus.script.interval import remap
from sonolus.script.runtime import time


class WatchHoldConnector(WatchArchetype):
    """A connector for hold notes."""

    name = "HoldConnector"

    first_ref: EntityRef[WatchNote] = imported()
    second_ref: EntityRef[WatchNote] = imported()

    end_time: float = imported()

    def spawn_time(self) -> float:
        return self.first.spawn_time()

    def despawn_time(self) -> float:
        return self.second.target_scaled_time

    def update_sequential(self):
        if self.first.target_time <= time() < self.second.target_time and self.head.has_active_touch:
            self.head.hold_lane = remap(self.first.y, self.second.y, self.first.lane, self.second.lane, 0)

    def update_parallel(self):
        draw_hold_connector(
            self.first.lane,
            self.second.lane,
            self.first.y,
            self.second.y,
        )

    @property
    def first(self):
        return self.first_ref.get()

    @property
    def second(self):
        return self.second_ref.get()

    @property
    def head(self):
        return self.first.head


class WatchSimLine(WatchArchetype):
    """A line connecting two simultaneous notes."""

    name = "SimLine"

    first_ref: EntityRef[WatchNote] = imported()
    second_ref: EntityRef[WatchNote] = imported()

    def spawn_time(self) -> float:
        return self.first.spawn_time()

    def despawn_time(self) -> float:
        return min(self.first.despawn_time(), self.second.despawn_time())

    def update_parallel(self):
        draw_sim_line(
            self.first.lane,
            self.second.lane,
            self.first.y,
        )

    @property
    def first(self):
        return self.first_ref.get()

    @property
    def second(self):
        return self.second_ref.get()



from __future__ import annotations

import json
from collections.abc import Callable
from os import PathLike
from pathlib import Path
from typing import Any, Literal

from sonolus.build.collection import Asset, load_asset
from sonolus.script.archetype import PlayArchetype, PreviewArchetype, WatchArchetype, _BaseArchetype
from sonolus.script.bucket import Buckets, EmptyBuckets
from sonolus.script.effect import Effects, EmptyEffects
from sonolus.script.instruction import (
    EmptyInstructionIcons,
    EmptyInstructions,
    TutorialInstructionIcons,
    TutorialInstructions,
)
from sonolus.script.metadata import AnyText, Tag, as_localization_text
from sonolus.script.options import EmptyOptions, Options
from sonolus.script.particle import EmptyParticles, Particles
from sonolus.script.sprite import EmptySkin, Skin
from sonolus.script.ui import UiConfig


class ExportedEngine:
    """An exported Sonolus.py engine."""

    def __init__(
        self,
        *,
        item: dict,
        thumbnail: bytes,
        play_data: bytes,
        watch_data: bytes,
        preview_data: bytes,
        tutorial_data: bytes,
        rom: bytes | None = None,
        configuration: bytes,
    ) -> None:
        self.item = item
        self.thumbnail = thumbnail
        self.play_data = play_data
        self.watch_data = watch_data
        self.preview_data = preview_data
        self.tutorial_data = tutorial_data
        self.rom = rom
        self.configuration = configuration

    def write_to_dir(self, path: PathLike):
        """Write the exported engine to a directory."""
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        (path / "item.json").write_text(json.dumps(self.item, ensure_ascii=False), encoding="utf-8")
        (path / "thumbnail").write_bytes(self.thumbnail)
        (path / "playData").write_bytes(self.play_data)
        (path / "watchData").write_bytes(self.watch_data)
        (path / "previewData").write_bytes(self.preview_data)
        (path / "tutorialData").write_bytes(self.tutorial_data)
        if self.rom is not None:
            (path / "rom").write_bytes(self.rom)
        (path / "configuration").write_bytes(self.configuration)


class Engine:
    """A Sonolus.py engine.

    Args:
        name: The name of the engine.
        title: The title of the engine.
        subtitle: The subtitle of the engine.
        author: The author of the engine.
        skin: The default skin for the engine.
        background: The default background for the engine.
        effect: The default effect for the engine.
        particle: The default particle for the engine.
        thumbnail: The thumbnail for the engine.
        data: The engine's modes and configurations.
        tags: The tags of the engine.
        description: The description of the engine.
        meta: Additional metadata of the engine.
    """

    version: Literal[13] = 13

    def __init__(
        self,
        *,
        name: str,
        title: AnyText | None = None,
        subtitle: AnyText = "Sonolus.py Engine",
        author: AnyText = "Unknown",
        skin: str | None = None,
        background: str | None = None,
        effect: str | None = None,
        particle: str | None = None,
        thumbnail: Asset | None = None,
        data: EngineData,
        tags: list[Tag] | None = None,
        description: AnyText | None = None,
        meta: Any = None,
    ) -> None:
        self.name = name
        self.title = as_localization_text(title or name)
        self.subtitle = as_localization_text(subtitle)
        self.author = as_localization_text(author)
        self.skin = skin
        self.background = background
        self.effect = effect
        self.particle = particle
        self.thumbnail = thumbnail
        self.data = data
        self.tags = tags or []
        self.description = as_localization_text(description) if description is not None else None
        self.meta = meta

    def export(self) -> ExportedEngine:
        """Export the engine in a sonolus-pack compatible format.

        Returns:
            An exported engine.
        """
        from sonolus.build.engine import package_engine
        from sonolus.build.project import BLANK_PNG

        item = {
            "version": self.version,
            "title": self.title,
            "subtitle": self.subtitle,
            "author": self.author,
            "tags": [tag.as_dict() for tag in self.tags],
            "skin": self.skin,
            "background": self.background,
            "effect": self.effect,
            "particle": self.particle,
        }
        packaged = package_engine(self.data)
        if self.description is not None:
            item["description"] = self.description
        if self.meta is not None:
            item["meta"] = self.meta
        return ExportedEngine(
            item=item,
            thumbnail=load_asset(self.thumbnail) if self.thumbnail is not None else BLANK_PNG,
            play_data=packaged.play_data,
            watch_data=packaged.watch_data,
            preview_data=packaged.preview_data,
            tutorial_data=packaged.tutorial_data,
            rom=packaged.rom,
            configuration=packaged.configuration,
        )


def default_callback() -> Any:
    return 0.0


class PlayMode:
    """A play mode definition.

    Args:
        archetypes: A list of play archetypes.
        skin: The skin for the play mode.
        effects: The effects for the play mode.
        particles: The particles for the play mode.
        buckets: The buckets for the play mode.
    """

    def __init__(
        self,
        *,
        archetypes: list[type[_BaseArchetype]] | None = None,
        skin: Skin = EmptySkin,
        effects: Effects = EmptyEffects,
        particles: Particles = EmptyParticles,
        buckets: Buckets = EmptyBuckets,
    ) -> None:
        self.archetypes = archetypes or []
        self.skin = skin
        self.effects = effects
        self.particles = particles
        self.buckets = buckets

        for archetype in self.archetypes:
            if not issubclass(archetype, PlayArchetype):
                raise ValueError(f"archetype {archetype} is not a PlayArchetype")


class WatchMode:
    """A watch mode definition.

    Args:
        archetypes: A list of watch archetypes.
        skin: The skin for the watch mode.
        effects: The effects for the watch mode.
        particles: The particles for the watch mode.
        buckets: The buckets for the watch mode.
        update_spawn: A callback returning the spawn time used by archetypes.
    """

    def __init__(
        self,
        *,
        archetypes: list[type[_BaseArchetype]] | None = None,
        skin: Skin = EmptySkin,
        effects: Effects = EmptyEffects,
        particles: Particles = EmptyParticles,
        buckets: Buckets = EmptyBuckets,
        update_spawn: Callable[[], float],
    ) -> None:
        self.archetypes = archetypes or []
        self.skin = skin
        self.effects = effects
        self.particles = particles
        self.buckets = buckets
        self.update_spawn = update_spawn

        for archetype in self.archetypes:
            if not issubclass(archetype, WatchArchetype):
                raise ValueError(f"archetype {archetype} is not a WatchArchetype")


class PreviewMode:
    """A preview mode definition.

    Args:
        archetypes: A list of preview archetypes.
        skin: The skin for the preview mode.
    """

    def __init__(
        self,
        *,
        archetypes: list[type[_BaseArchetype]] | None = None,
        skin: Skin = EmptySkin,
    ) -> None:
        self.archetypes = archetypes or []
        self.skin = skin

        for archetype in self.archetypes:
            if not issubclass(archetype, PreviewArchetype):
                raise ValueError(f"archetype {archetype} is not a PreviewArchetype")


class TutorialMode:
    """A tutorial mode definition.

    Args:
        skin: The skin for the tutorial mode.
        effects: The effects for the tutorial mode.
        particles: The particles for the tutorial mode.
        instructions: The instructions for the tutorial mode.
        instruction_icons: The instruction icons for the tutorial mode.
        preprocess: A callback to be called before the tutorial starts.
        navigate: A callback to be called when the user navigates.
        update: A callback to be called each frame.
    """

    def __init__(
        self,
        *,
        skin: Skin = EmptySkin,
        effects: Effects = EmptyEffects,
        particles: Particles = EmptyParticles,
        instructions: TutorialInstructions = EmptyInstructions,
        instruction_icons: TutorialInstructionIcons = EmptyInstructionIcons,
        preprocess: Callable[[], None],
        navigate: Callable[[], None],
        update: Callable[[], None],
    ) -> None:
        self.skin = skin
        self.effects = effects
        self.particles = particles
        self.instructions = instructions
        self.instruction_icons = instruction_icons
        self.preprocess = preprocess
        self.navigate = navigate
        self.update = update


def empty_play_mode() -> PlayMode:
    """Create an empty play mode."""
    return PlayMode()


def empty_watch_mode() -> WatchMode:
    """Create an empty watch mode."""
    return WatchMode(update_spawn=default_callback)


def empty_preview_mode() -> PreviewMode:
    """Create an empty preview mode."""
    return PreviewMode()


def empty_tutorial_mode() -> TutorialMode:
    """Create an empty tutorial mode."""
    return TutorialMode(
        preprocess=default_callback,
        navigate=default_callback,
        update=default_callback,
    )


class EngineData:
    """A Sonolus.py engine's modes and configurations.

    Args:
        ui: The UI configuration.
        options: The options for the engine.
        play: The play mode configuration.
        watch: The watch mode configuration.
        preview: The preview mode configuration.
        tutorial: The tutorial mode configuration.
    """

    def __init__(
        self,
        *,
        ui: UiConfig | None = None,
        options: Options = EmptyOptions,
        play: PlayMode | None = None,
        watch: WatchMode | None = None,
        preview: PreviewMode | None = None,
        tutorial: TutorialMode | None = None,
    ) -> None:
        self.ui = ui or UiConfig()
        self.options = options
        self.play = play or empty_play_mode()
        self.watch = watch or empty_watch_mode()
        self.preview = preview or empty_preview_mode()
        self.tutorial = tutorial or empty_tutorial_mode()



from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated, Any, NewType, dataclass_transform, get_origin

from sonolus.backend.ops import Op
from sonolus.script.internal.introspection import get_field_specifiers
from sonolus.script.internal.native import native_function
from sonolus.script.record import Record


class Effect(Record):
    """Sound effect clip.

    Usage:
        ```python
        Effect(id: int)
        ```
    """

    id: int
    """Effect ID."""

    @property
    def is_available(self) -> bool:
        """Return whether the effect clip is available."""
        return _has_effect_clip(self.id)

    def play(self, distance: float = 0) -> None:
        """Play the effect clip.

        If the clip was already played within the specified distance, it will be skipped.

        Arguments:
            distance: Minimum time in seconds since the last play for the effect to play.
        """
        _play(self.id, distance)

    def schedule(self, time: float, distance: float = 0) -> None:
        """Schedule the effect clip to play at a specific time.

        This is not suitable for real-time effects such as responses to user input.
        Use [`play`][sonolus.script.effect.Effect.play] instead.

        This may be called in preprocess to schedule effects upfront.

        If the clip would play within the specified distance of another play, it will be skipped.

        Arguments:
            time: Time in seconds when the effect should play.
            distance: Minimum time in seconds after a previous play for the effect to play.
        """
        _play_scheduled(self.id, time, distance)

    def loop(self) -> LoopedEffectHandle:
        """Play the effect clip in a loop until stopped.

        Returns:
            A handle to stop the loop.
        """
        return LoopedEffectHandle(_play_looped(self.id))

    def schedule_loop(self, start_time: float) -> ScheduledLoopedEffectHandle:
        """Schedule the effect clip to play in a loop until stopped.

        This is not suitable for real-time effects such as responses to user input.
        Use [`loop`][sonolus.script.effect.Effect.loop] instead.

        Returns:
            A handle to stop the loop.
        """
        return ScheduledLoopedEffectHandle(_play_looped_scheduled(self.id, start_time))


class LoopedEffectHandle(Record):
    """Handle to stop a looped effect."""

    id: int

    def stop(self) -> None:
        """Stop the looped effect."""
        _stop_looped(self.id)


class ScheduledLoopedEffectHandle(Record):
    """Handle to stop a scheduled looped effect."""

    id: int

    def stop(self, end_time: float) -> None:
        """Stop the scheduled looped effect."""
        _stop_looped_scheduled(self.id, end_time)


@native_function(Op.HasEffectClip)
def _has_effect_clip(effect_id: int) -> bool:
    raise NotImplementedError


@native_function(Op.Play)
def _play(effect_id: int, distance: float) -> None:
    raise NotImplementedError


@native_function(Op.PlayLooped)
def _play_looped(effect_id: int) -> int:
    raise NotImplementedError


@native_function(Op.PlayLoopedScheduled)
def _play_looped_scheduled(effect_id: int, start_time: float) -> int:
    raise NotImplementedError


@native_function(Op.PlayScheduled)
def _play_scheduled(effect_id: int, time: float, distance: float) -> None:
    raise NotImplementedError


@native_function(Op.StopLooped)
def _stop_looped(handle: int) -> None:
    raise NotImplementedError


@native_function(Op.StopLoopedScheduled)
def _stop_looped_scheduled(handle: int, end_time: float) -> None:
    raise NotImplementedError


@dataclass
class EffectInfo:
    name: str


def effect(name: str) -> Any:
    """Define a sound effect clip with the given name."""
    return EffectInfo(name)


type Effects = NewType("Effects", Any)  # type: ignore


@dataclass_transform(kw_only_default=True)
def effects[T](cls: type[T]) -> T | Effects:
    """Decorator to define effect clips.

    Usage:
        ```python
        @effects
        class Effects:
            miss: StandardEffect.MISS
            other: Effect = effect("other")
        ```
    """
    if len(cls.__bases__) != 1:
        raise ValueError("Effects class must not inherit from any class (except object)")
    instance = cls()
    names = []
    for i, (name, annotation) in enumerate(get_field_specifiers(cls).items()):
        if get_origin(annotation) is not Annotated:
            raise TypeError(f"Invalid annotation for effects: {annotation}")
        annotation_type = annotation.__args__[0]
        annotation_values = annotation.__metadata__
        if annotation_type is not Effect:
            raise TypeError(f"Invalid annotation for effects: {annotation}, expected annotation of type Effect")
        if len(annotation_values) != 1 or not isinstance(annotation_values[0], EffectInfo):
            raise TypeError(f"Invalid annotation for effects: {annotation}, expected a single string annotation value")
        effect_name = annotation_values[0].name
        names.append(effect_name)
        setattr(instance, name, Effect(i))
    instance._effects_ = names
    instance._is_comptime_value_ = True
    return instance


class StandardEffect:
    """Standard sound effect clips."""

    MISS = Annotated[Effect, effect("#MISS")]
    PERFECT = Annotated[Effect, effect("#PERFECT")]
    GREAT = Annotated[Effect, effect("#GREAT")]
    GOOD = Annotated[Effect, effect("#GOOD")]
    HOLD = Annotated[Effect, effect("#HOLD")]
    MISS_ALTERNATIVE = Annotated[Effect, effect("#MISS_ALTERNATIVE")]
    PERFECT_ALTERNATIVE = Annotated[Effect, effect("#PERFECT_ALTERNATIVE")]
    GREAT_ALTERNATIVE = Annotated[Effect, effect("#GREAT_ALTERNATIVE")]
    GOOD_ALTERNATIVE = Annotated[Effect, effect("#GOOD_ALTERNATIVE")]
    HOLD_ALTERNATIVE = Annotated[Effect, effect("#HOLD_ALTERNATIVE")]
    STAGE = Annotated[Effect, effect("#STAGE")]


@effects
class EmptyEffects:
    pass



from __future__ import annotations

from math import atan2, cos, sin

from sonolus.script.array import Array
from sonolus.script.array_like import ArrayLike
from sonolus.script.num import Num
from sonolus.script.record import Record


class Vec2(Record):
    """A 2D vector.

    Usage:
        ```python
        Vec2(x: float, y: float)
        ```
    """

    x: float
    y: float

    @classmethod
    def zero(cls) -> Vec2:
        """Return a vector with x and y set to 0.

        Returns:
            A new vector with x=0 and y=0.
        """
        return cls(x=0, y=0)

    @classmethod
    def one(cls) -> Vec2:
        """Return a vector with x and y set to 1.

        Returns:
            A new vector with x=1 and y=1.
        """
        return cls(x=1, y=1)

    @classmethod
    def up(cls) -> Vec2:
        """Return a vector pointing upwards (x=0, y=1).

        Returns:
            A new vector pointing upwards.
        """
        return cls(x=0, y=1)

    @classmethod
    def down(cls) -> Vec2:
        """Return a vector pointing downwards (x=0, y=-1).

        Returns:
            A new vector pointing downwards.
        """
        return cls(x=0, y=-1)

    @classmethod
    def left(cls) -> Vec2:
        """Return a vector pointing to the left (x=-1, y=0).

        Returns:
            A new vector pointing to the left.
        """
        return cls(x=-1, y=0)

    @classmethod
    def right(cls) -> Vec2:
        """Return a vector pointing to the right (x=1, y=0).

        Returns:
            A new vector pointing to the right.
        """
        return cls(x=1, y=0)

    @classmethod
    def unit(cls, angle: float) -> Vec2:
        """Return a unit vector (magnitude 1) at a given angle in radians.

        Args:
            angle: The angle in radians.

        Returns:
            A new unit vector at the specified angle.
        """
        return Vec2(x=cos(angle), y=sin(angle))

    @property
    def magnitude(self) -> float:
        """Calculate the magnitude (length) of the vector.

        Returns:
            The magnitude of the vector.
        """
        return (self.x**2 + self.y**2) ** 0.5

    @property
    def angle(self) -> float:
        """Calculate the angle of the vector in radians from the positive x-axis.

        Returns:
            The angle of the vector in radians.
        """
        return atan2(self.y, self.x)

    def dot(self, other: Vec2) -> float:
        """Calculate the dot product of this vector with another vector.

        Args:
            other: The other vector to calculate the dot product with.

        Returns:
            The dot product of the two vectors.
        """
        return self.x * other.x + self.y * other.y

    def rotate(self, angle: float) -> Vec2:
        """Rotate the vector by a given angle in radians and return a new vector.

        Args:
            angle: The angle to rotate the vector by, in radians. Positive angles rotate counterclockwise.

        Returns:
            A new vector rotated by the given angle.
        """
        return Vec2(
            x=self.x * cos(angle) - self.y * sin(angle),
            y=self.x * sin(angle) + self.y * cos(angle),
        )

    def rotate_about(self, angle: float, pivot: Vec2) -> Vec2:
        """Rotate the vector about a pivot by a given angle in radians and return a new vector.

        Args:
            angle: The angle to rotate the vector by, in radians. Positive angles rotate counterclockwise.
            pivot: The pivot point to rotate about.

        Returns:
            A new vector rotated about the pivot by the given angle.
        """
        return (self - pivot).rotate(angle) + pivot

    def normalize(self) -> Vec2:
        """Normalize the vector (set the magnitude to 1) and return a new vector.

        Returns:
            A new vector with magnitude 1.
        """
        magnitude = self.magnitude
        return Vec2(x=self.x / magnitude, y=self.y / magnitude)

    def orthogonal(self) -> Vec2:
        """Return a vector orthogonal to this vector.

        The orthogonal vector is rotated 90 degrees counter-clockwise from this vector.

        Returns:
            A new vector orthogonal to this vector.
        """
        return Vec2(x=-self.y, y=self.x)

    @property
    def tuple(self) -> tuple[float, float]:
        """Return the vector as a tuple (x, y).

        Returns:
            A tuple representation of the vector.
        """
        return self.x, self.y

    def __add__(self, other: Vec2) -> Vec2:
        """Add this vector to another vector and return a new vector.

        Args:
            other: The vector to add.

        Returns:
            A new vector resulting from the addition.
        """
        return Vec2(x=self.x + other.x, y=self.y + other.y)

    def __sub__(self, other: Vec2) -> Vec2:
        """Subtract another vector from this vector and return a new vector.

        Args:
            other: The vector to subtract.

        Returns:
            A new vector resulting from the subtraction.
        """
        return Vec2(x=self.x - other.x, y=self.y - other.y)

    def __mul__(self, other: Vec2 | float) -> Vec2:
        """Multiply this vector by another vector or a scalar and return a new vector.

        Args:
            other: The vector or scalar to multiply by.

        Returns:
            A new vector resulting from the multiplication.
        """
        match other:
            case Vec2(x, y):
                return Vec2(x=self.x * x, y=self.y * y)
            case Num(factor):
                return Vec2(x=self.x * factor, y=self.y * factor)
            case _:
                return NotImplemented

    def __rmul__(self, other):
        match other:
            case Num(factor):
                return Vec2(x=self.x * factor, y=self.y * factor)
            case _:
                return NotImplemented

    def __truediv__(self, other: Vec2 | float) -> Vec2:
        """Divide this vector by another vector or a scalar and return a new vector.

        Args:
            other: The vector or scalar to divide by.

        Returns:
            A new vector resulting from the division.
        """
        match other:
            case Vec2(x, y):
                return Vec2(x=self.x / x, y=self.y / y)
            case Num(factor):
                return Vec2(x=self.x / factor, y=self.y / factor)
            case _:
                return NotImplemented

    def __neg__(self) -> Vec2:
        """Negate the vector (invert the direction) and return a new vector.

        Returns:
            A new vector with inverted direction.
        """
        return Vec2(x=-self.x, y=-self.y)


def pnpoly(vertices: ArrayLike[Vec2] | tuple[Vec2, ...], test: Vec2) -> bool:
    """Check if a point is inside a polygon.

    No guaranteed behavior for points on the edges or very close to the edges.

    Args:
        vertices: The vertices of the polygon.
        test: The point to test.

    Returns:
        Whether the point is inside the polygon.
    """
    if isinstance(vertices, tuple):
        vertices = Array(*vertices)
    i = 0
    j = len(vertices) - 1
    c = False
    while i < len(vertices):
        if (vertices[i].y > test.y) != (vertices[j].y > test.y) and test.x < (vertices[j].x - vertices[i].x) * (
            test.y - vertices[i].y
        ) / (vertices[j].y - vertices[i].y) + vertices[i].x:
            c = not c
        j = i
        i += 1
    return c
